"""Functions that can be used to test encodings."""
from typing import Any, Callable

from moptipy.api.encoding import Encoding, check_encoding
from moptipy.api.space import Space
from moptipy.tests.component import validate_component
from moptipy.utils.types import type_error


def validate_encoding(encoding: Encoding,
                      search_space: Space | None = None,
                      solution_space: Space | None = None,
                      make_search_space_element_valid:
                      Callable[[Any], Any] | None = lambda x: x,
                      is_deterministic: bool = True) -> None:
    """
    Check whether an object is a proper moptipy encoding.

    :param encoding: the encoding to test
    :param search_space: the search space
    :param make_search_space_element_valid: a method that can turn a point
        from the space into a valid point
    :param solution_space: the solution space
    :param is_deterministic: is the mapping deterministic?
    :raises ValueError: if `encoding` is not a valid
        :class:`~moptipy.api.encoding.Encoding`
    :raises TypeError: if `encoding` is of the wrong type or a wrong type is
        encountered
    """
    if not isinstance(encoding, Encoding):
        raise type_error(encoding, "encoding", Encoding)
    check_encoding(encoding)
    validate_component(encoding)

    count: int = 0
    if search_space is not None:
        count += 1
    if make_search_space_element_valid is not None:
        count += 1
    if solution_space is not None:
        count += 1
    if count <= 0:
        return
    if count < 3:
        raise ValueError(
            "either provide all of search_space, "
            "make_search_space_element_valid, and solution_space or none.")

    x1 = search_space.create()
    if x1 is None:
        raise ValueError("Provided search space created None?")
    x1 = make_search_space_element_valid(x1)
    search_space.validate(x1)

    y1 = solution_space.create()
    if y1 is None:
        raise ValueError("Provided solution space created None?")

    if not (hasattr(encoding, "decode") and callable(getattr(
            encoding, "decode"))):
        raise ValueError("encoding must have method decode.")

    encoding.decode(x1, y1)
    solution_space.validate(y1)
    s1: str = solution_space.to_str(y1)
    if s1 is None:
        raise ValueError("to_str() returned None!")
    if len(s1) <= 0:
        raise ValueError("to_str() return empty string")

    y2 = solution_space.create()
    if y2 is None:
        raise ValueError("Provided solution space created None?")
    if y1 is y2:
        raise ValueError("Provided solution space created "
                         "identical points?")
    encoding.decode(x1, y2)
    solution_space.validate(y2)
    s2: str = solution_space.to_str(y2)
    if s2 is None:
        raise ValueError("to_str() returned None!")
    if len(s2) <= 0:
        raise ValueError("to_str() return empty string")
    if is_deterministic:
        if not solution_space.is_equal(y1, y2):
            raise ValueError("Encoding must be deterministic and decode "
                             "identical points to same result.")
        if s1 != s2:
            raise ValueError(f"to_str(y1)={s1!r} but to_str(y2)={s2!r}!")

    x2 = search_space.create()
    if x2 is None:
        raise ValueError("Provided search space created None?")
    if x1 is x2:
        raise ValueError("Provided search space created "
                         "identical points?")

    search_space.copy(x2, x1)
    if not search_space.is_equal(x1, x2):
        raise ValueError("Copy method of search space did not result in "
                         "is_equal becoming true?")
    search_space.validate(x2)

    encoding.decode(x2, y2)
    solution_space.validate(y2)
    s2 = solution_space.to_str(y2)
    if s2 is None:
        raise ValueError("to_str() returned None!")
    if len(s2) <= 0:
        raise ValueError("to_str() return empty string")

    if is_deterministic:
        if not solution_space.is_equal(y1, y2):
            raise ValueError("Encoding must be deterministic and decode "
                             "equal points to same result.")
        if s1 != s2:
            raise ValueError(f"to_str(y1)={s1!r} but to_str(y2)={s2!r}!")



"""Plot a Gantt chart into one figure."""
from typing import Callable, Final, Iterable

from matplotlib.artist import Artist  # type: ignore
from matplotlib.axes import Axes  # type: ignore
from matplotlib.figure import Figure, SubplotBase  # type: ignore
from matplotlib.lines import Line2D  # type: ignore
from matplotlib.patches import Rectangle  # type: ignore
from matplotlib.text import Text  # type: ignore
from matplotlib.ticker import MaxNLocator  # type: ignore

import moptipy.utils.plot_defaults as pd
import moptipy.utils.plot_utils as pu
from moptipy.evaluation.axis_ranger import AxisRanger
from moptipy.examples.jssp.gantt import Gantt
from moptipy.examples.jssp.makespan import makespan
from moptipy.utils.lang import Lang
from moptipy.utils.types import type_error


def marker_lb(x: Gantt) -> tuple[str, int | float]:
    """
    Compute the marker for the lower bound.

    :param x: the Gantt chart
    :return: the lower bound marker
    """
    return Lang.current()["lower_bound_short"],\
        x.instance.makespan_lower_bound


def marker_makespan(x: Gantt) -> tuple[str, int | float]:
    """
    Compute the marker for the makespan.

    :param x: the Gantt chart
    :return: the makespan marker
    """
    return Lang.current()["makespan"], makespan(x)


#: the color for markers at the left end
__LEFT_END_MARK: Final[tuple[float, float, float]] = (0.95, 0.02, 0.02)
#: the color for markers at the right end
__RIGHT_END_MARK: Final[tuple[float, float, float]] = (0.02, 0.02, 0.95)
#: the color for markers in the middle
__MIDDLE_MARK: Final[tuple[float, float, float]] = pd.COLOR_BLACK


def plot_gantt_chart(
        gantt: Gantt | str,
        figure: SubplotBase | Figure,
        markers: Iterable[tuple[str, int | float] | Callable[
            [Gantt], tuple[str, int | float]]] | None = (marker_lb,),
        x_axis: AxisRanger | Callable[[Gantt], AxisRanger] =
        lambda gantt: AxisRanger(chosen_min=0),
        importance_to_line_width_func: Callable[[int], float] =
        pd.importance_to_line_width,
        importance_to_font_size_func: Callable[[int], float] =
        pd.importance_to_font_size,
        info: None | str | Callable[[Gantt], str] =
        lambda gantt: Lang.current().format_str("gantt_info", gantt=gantt),
        x_grid: bool = False,
        y_grid: bool = False,
        x_label: None | str | Callable[[Gantt], str] =
        Lang.translate_call("time"),
        x_label_inside: bool = True,
        x_label_location: float = 1.0,
        y_label: None | str | Callable[[Gantt], str] =
        Lang.translate_call("machine"),
        y_label_inside: bool = True,
        y_label_location: float = 0.5) -> Axes:
    """
    Plot a Gantt chart.

    :param gantt: the gantt chart or a path to a file to load it from
    :param figure: the figure
    :param markers: a set of markers
    :param x_axis: the ranger for the x-axis
    :param info: the optional info header
    :param importance_to_font_size_func: the function converting
        importance values to font sizes
    :param importance_to_line_width_func: the function converting
        importance values to line widths
    :param x_grid: should we have a grid along the x-axis?
    :param y_grid: should we have a grid along the y-axis?
    :param x_label: a callable returning the label for
        the x-axis, a label string, or `None` if no label should be put
    :param x_label_inside: put the x-axis label inside the plot (so that
        it does not consume additional vertical space)
    :param x_label_location: the location of the x-label
    :param y_label: a callable returning the label for
        the y-axis, a label string, or `None` if no label should be put
    :param y_label_inside: put the y-axis label inside the plot (so that
        it does not consume additional horizontal space)
    :param y_label_location: the location of the y-label
    :returns: the axes object to allow you to add further plot elements
    """
    if isinstance(gantt, str):
        gantt = Gantt.from_log(gantt)
    if not isinstance(gantt, Gantt):
        raise type_error(gantt, "gantt", (Gantt, str))
    axes: Final[Axes] = pu.get_axes(figure)

    # grab the data
    jobs: Final[int] = gantt.instance.jobs
    machines: Final[int] = gantt.instance.machines

    # Set up the x-axis range.
    if callable(x_axis):
        x_axis = x_axis(gantt)
    if not isinstance(x_axis, AxisRanger):
        raise type_error(x_axis, "x_axis", AxisRanger)

    # Compute all the marks
    marks: dict[int | float, str] = {}
    if markers is not None:
        if not isinstance(markers, Iterable):
            raise type_error(markers, "markers", Iterable)
        for marker in markers:
            if callable(marker):
                marker = marker(gantt)
            if not marker:
                continue
            if isinstance(marker, tuple):
                name, val = marker
                if (not name) or (not val):
                    continue
            else:
                raise type_error(marker, "marker", tuple, True)
            if not isinstance(name, str):
                raise type_error(name, "marker name", str)
            if not isinstance(val, (int, float)):
                raise type_error(val, "marker", (int, float))
            if val in marks:
                marks[val] = f"{marks[val]}/{name}"
            else:
                marks[val] = name
                x_axis.register_value(val)

    # Add x-axis data
    x_axis.register_array(gantt[:, 0, 1].flatten())  # register start times
    x_axis.register_array(gantt[:, -1, 2].flatten())  # register end times
    x_axis.apply(axes, "x")
    xmin, xmax = axes.get_xlim()

    # Set up the y-axis range.
    height: Final[float] = 0.7
    bar_ofs: Final[float] = height / 2
    y_min: Final[float] = -((7 * bar_ofs) / 6)
    y_max: Final[float] = (machines - 1) + ((7 * bar_ofs) / 6)

    axes.set_ylim(y_min, y_max)
    axes.set_ybound(y_min, y_max)
    axes.yaxis.set_major_locator(MaxNLocator(nbins="auto",
                                             integer=True))

    # get the color and font styles
    colors: Final[tuple] = pd.distinct_colors(jobs)
    font_size: Final[float] = importance_to_font_size_func(-1)

    # get the transforms needed to obtain text dimensions
    rend: Final = pu.get_renderer(axes)
    inv: Final = axes.transData.inverted()

    # draw the grid
    if x_grid or y_grid:
        grid_lwd = importance_to_line_width_func(-1)
        if x_grid:
            axes.grid(axis="x", color=pd.GRID_COLOR, linewidth=grid_lwd)
        if y_grid:
            axes.grid(axis="y", color=pd.GRID_COLOR, linewidth=grid_lwd)

    z_order: int = 0

    # print the marker lines
    for val, _ in marks.items():
        axes.add_artist(Line2D(xdata=(val, val),
                               ydata=(y_min, y_max),
                               color=__LEFT_END_MARK if val <= xmin
                               else __RIGHT_END_MARK if val >= xmax
                               else __MIDDLE_MARK,
                               linewidth=2.0,
                               zorder=z_order))
        z_order += 1

    # plot the jobs
    for machine in range(machines):
        for jobi in range(jobs):
            job, x_start, x_end = gantt[machine, jobi, :]

            if x_end <= x_start:  # skip operations that take 0 time
                continue

            background = colors[job]
            foreground = pd.text_color_for_background(colors[job])
            jobstr = str(job)

            # first plot the colored rectangle
            y_start = machine - bar_ofs

            axes.add_artist(Rectangle(
                xy=(x_start, y_start),
                width=(x_end - x_start),
                height=height,
                color=background,
                linewidth=0,
                zorder=z_order))
            z_order += 1

            # Now we insert the job IDs, which is a bit tricky:
            # First, the rectangle may be too small to hold the text.
            # So we need to get the text bounding box size to compare it with
            # the rectangle's size.
            # If that fits, we can print the text.
            # Second, the printed text tends to be slightly off vertically,
            # even if we try to center it vertically. This is because fonts
            # can extend below their baseline which seems to be considered
            # during vertical alignment although job IDs are numbers and that
            # does not apply here. Therefore, we try to re-adjust the boxes
            # in a very, very crude way.
            xp: float = 0.5 * (x_start + x_end)
            yp = machine

            # Get the size of the text using a temporary text
            # that gets immediately deleted again.
            tmp: Text = axes.text(x=xp, y=yp, s=jobstr,
                                  fontsize=font_size,
                                  color=foreground,
                                  horizontalalignment="center",
                                  verticalalignment="baseline")
            bb_bl = inv.transform_bbox(tmp.get_window_extent(
                renderer=rend))
            Artist.set_visible(tmp, False)
            Artist.remove(tmp)
            del tmp

            if (bb_bl.width < 0.97 * (x_end - x_start)) and \
                    (bb_bl.height < (0.97 * height)):
                # OK, there is enough space. Let's re-compute the y
                # offset to do proper alignment using another temporary
                # text.
                tmp = axes.text(x=xp, y=yp, s=jobstr,
                                fontsize=font_size,
                                color=foreground,
                                horizontalalignment="center",
                                verticalalignment="bottom")
                bb_bt = inv.transform_bbox(tmp.get_window_extent(
                    renderer=rend))
                Artist.set_visible(tmp, False)
                Artist.remove(tmp)
                del tmp

                # Now we can really print the actual text with a more or less
                # nice vertical alignment.
                adj = bb_bl.y0 - bb_bt.y0
                if adj < 0:
                    yp += adj / 3

                axes.text(x=xp, y=yp, s=jobstr,
                          fontsize=font_size,
                          color=foreground,
                          horizontalalignment="center",
                          verticalalignment="center",
                          zorder=z_order)
                z_order += 1

    # print the marker labels
    bbox = {"boxstyle": "round",
            "color": "white",
            "fill": True,
            "linewidth": 0,
            "alpha": 0.9}
    y_mark: Final[float] = -0.1  # machines - 1 + (0.9 * bar_ofs) for top
    for val, name in marks.items():
        axes.annotate(text=f"{name}={val}",
                      xy=(val, y_mark),
                      xytext=(-4, -4),
                      verticalalignment="bottom",
                      horizontalalignment="right",
                      xycoords="data",
                      textcoords="offset points",
                      fontsize=font_size,
                      color=__LEFT_END_MARK if val <= xmin
                      else __RIGHT_END_MARK if val >= xmax
                      else __MIDDLE_MARK,
                      rotation=90,
                      bbox=bbox,
                      zorder=z_order)
        z_order += 1

    info_font_size: Final[float] = importance_to_font_size_func(0)
    pu.label_axes(axes=axes,
                  x_label=x_label(gantt) if callable(x_label) else x_label,
                  x_label_inside=x_label_inside,
                  x_label_location=x_label_location,
                  y_label=y_label(gantt) if callable(y_label) else y_label,
                  y_label_inside=y_label_inside,
                  y_label_location=y_label_location,
                  font_size=info_font_size,
                  z_order=z_order)
    z_order = z_order + 1

    if callable(info):
        info = info(gantt)
    if info is not None:
        if not isinstance(info, str):
            raise type_error(info, "info", str)
        pu.label_box(axes=axes,
                     text=info,
                     x=0.5,
                     y=1,
                     font_size=importance_to_font_size_func(1),
                     may_rotate_text=False,
                     z_order=z_order)
    return axes



"""A utility to specify axis ranges."""
import sys
from math import inf, isfinite
from typing import Callable, Final

import numpy as np
from matplotlib.axes import Axes  # type: ignore

from moptipy.api.logging import (
    KEY_LAST_IMPROVEMENT_FE,
    KEY_LAST_IMPROVEMENT_TIME_MILLIS,
    KEY_TOTAL_FES,
    KEY_TOTAL_TIME_MILLIS,
)
from moptipy.evaluation.base import (
    F_NAME_NORMALIZED,
    F_NAME_RAW,
    F_NAME_SCALED,
    TIME_UNIT_FES,
    TIME_UNIT_MILLIS,
)
from moptipy.evaluation.end_statistics import KEY_ERT_FES, KEY_ERT_TIME_MILLIS
from moptipy.utils.types import type_error

#: The internal minimum float value for log-scaled axes.
_MIN_LOG_FLOAT: Final[float] = sys.float_info.min


class AxisRanger:
    """An object for simplifying axis range computations."""

    def __init__(self,
                 chosen_min: float | None = None,
                 chosen_max: float | None = None,
                 use_data_min: bool = True,
                 use_data_max: bool = True,
                 log_scale: bool = False,
                 log_base: float | None = None):
        """
        Initialize the axis ranger.

        :param chosen_min: the chosen minimum
        :param chosen_max: the chosen maximum
        :param use_data_min: use the minimum found in the data?
        :param use_data_max:  use the maximum found in the data?
        :param log_scale: should the axis be log-scaled?
        :param log_base: the base to be used for the logarithm
        """
        if not isinstance(log_scale, bool):
            raise type_error(log_scale, "log_scale", bool)
        #: Should the axis be log-scaled?
        self.log_scale: Final[bool] = log_scale

        self.__log_base: Final[float | None] = \
            log_base if self.log_scale else None
        if self.__log_base is not None:
            if not isinstance(log_base, float):
                raise type_error(log_base, "log_base", float)
            if log_base <= 1.0:
                raise ValueError(f"log_base must be > 1, but is {log_base}.")

        if chosen_min is not None:
            if not isinstance(chosen_min, (float, int)):
                raise type_error(chosen_min, "chosen_min", (int, float))
            chosen_min = float(chosen_min)
            if not isfinite(chosen_min):
                raise ValueError(f"chosen_min cannot be {chosen_min}.")
            if self.log_scale and (chosen_min <= 0):
                raise ValueError(
                    f"if log_scale={self.log_scale}, then chosen_min must "
                    f"be > 0, but is {chosen_min}.")

        #: The pre-defined, chosen minimum axis value.
        self.__chosen_min: Final[float | None] = chosen_min

        if chosen_max is not None:
            if not isinstance(chosen_max, (float, int)):
                raise type_error(chosen_max, "chosen_max", (int, float))
            chosen_max = float(chosen_max)
            if not isfinite(chosen_max):
                raise ValueError(f"chosen_max cannot be {chosen_max}.")
            if (self.__chosen_min is not None) and \
                    (chosen_max <= self.__chosen_min):
                raise ValueError(f"If chosen_min is {self.__chosen_min}, then"
                                 f" chosen_max cannot be {chosen_max}.")

        #: The pre-defined, chosen maximum axis value.
        self.__chosen_max: Final[float | None] = chosen_max

        if not isinstance(use_data_min, bool):
            raise type_error(use_data_min, "use_data_min", bool)
        #: Should we use the data min value?
        self.__use_data_min: Final[bool] = use_data_min

        if not isinstance(use_data_max, bool):
            raise type_error(use_data_max, "use_data_max", bool)
        #: Should we use the data max value?
        self.__use_data_max: Final[bool] = use_data_max

        #: The minimum detected from the data.
        self.__detected_min: float = inf

        #: The maximum detected from the data.
        self.__detected_max: float = _MIN_LOG_FLOAT if self.log_scale \
            else -inf

        #: Did we detect a minimum?
        self.__has_detected_min = False
        #: Did we detect a maximum?
        self.__has_detected_max = False

    def register_array(self, data: np.ndarray) -> None:
        """
        Register a data array.

        :param data: the data to register
        """
        if self.__use_data_min or self.__use_data_max:
            d = data[np.isfinite(data)]
            if self.__use_data_min:
                self.register_value(float(d.min()))
            if self.__use_data_max:
                self.register_value(float(d.max()))

    def register_value(self, value: float) -> None:
        """
        Register a single value.

        :param value: the data to register
        """
        if isfinite(value):
            if self.__use_data_min and (
                    (value < self.__detected_min)
                    and ((value > 0.0) or (not self.log_scale))):
                self.__detected_min = value
                self.__has_detected_min = True
            if self.__use_data_max and (value > self.__detected_max):
                self.__detected_max = value
                self.__has_detected_max = True

    def apply(self, axes: Axes, which_axis: str) -> None:
        """
        Apply this axis ranger to the given axis.

        :param axes: the axes object to which the ranger shall be applied
        :param which_axis: the axis to which it should be applied, either
            `"x"` or `"y"` or both (`"xy"`)
        """
        if not isinstance(which_axis, str):
            raise type_error(which_axis, "which_axis", str)

        for is_x_axis in (True, False):
            if ("x" if is_x_axis else "y") not in which_axis:
                continue

            use_min, use_max = \
                axes.get_xlim() if is_x_axis else axes.get_ylim()

            if not isfinite(use_min):
                raise ValueError(f"Minimum data interval cannot be {use_min}.")
            if not isfinite(use_max):
                raise ValueError(f"Maximum data interval cannot be {use_max}.")
            if use_max <= use_min:
                raise ValueError(f"Invalid axis range[{use_min},{use_max}].")

            replace_range = False

            if self.__chosen_min is not None:
                use_min = self.__chosen_min
                replace_range = True
            elif self.__use_data_min:
                if not self.__has_detected_min:
                    raise ValueError("No minimum in data detected.")
                use_min = self.__detected_min
                replace_range = True

            if self.__chosen_max is not None:
                use_max = self.__chosen_max
                replace_range = True
            elif self.__use_data_max:
                if not self.__has_detected_max:
                    raise ValueError("No maximum in data detected.")
                use_max = self.__detected_max
                replace_range = True

            if replace_range:
                if use_min >= use_max:
                    raise ValueError(
                        f"Invalid computed range [{use_min},{use_max}].")
                if is_x_axis:
                    axes.set_xlim(use_min, use_max)
                else:
                    axes.set_ylim(use_min, use_max)

            if self.log_scale:
                if use_min <= 0:
                    raise ValueError("minimum must be positive if log scale "
                                     f"is defined, but found {use_min}.")
                if is_x_axis:
                    if self.__log_base is None:
                        axes.semilogx()
                    else:
                        axes.semilogx(base=self.__log_base)
                else:
                    if self.__log_base is None:
                        axes.semilogy()
                    else:
                        axes.semilogy(base=self.__log_base)

    def get_pinf_replacement(self) -> float:
        """
        Get a reasonable finite value that can replace positive infinity.

        :return: a reasonable finite value that can be used to replace
            positive infinity
        """
        data_max: float = 0.0
        if self.__chosen_max is not None:
            data_max = self.__chosen_max
        elif self.__detected_max is not None:
            data_max = self.__detected_max
        return min(1e100, max(1e70, 1e5 * data_max))

    def get_0_replacement(self) -> float:
        """
        Get a reasonable positive finite value that can replace `0`.

        :return: a reasonable finite value that can be used to replace
            `0`
        """
        data_min: float = 1e-100
        if self.__chosen_min is not None:
            data_min = self.__chosen_min
        elif self.__detected_min is not None:
            data_min = self.__detected_min
        return max(1e-100, min(1e-70, 1e-5 * data_min))

    @staticmethod
    def for_axis(name: str,
                 chosen_min: float | None = None,
                 chosen_max: float | None = None,
                 use_data_min: bool | None = None,
                 use_data_max: bool | None = None,
                 log_scale: bool | None = None,
                 log_base: float | None = None) -> "AxisRanger":
        """
        Create a default axis ranger based on the axis type.

        The axis ranger will use the minimal values and log scaling options
        that usually make sense for the dimension, unless overridden by the
        optional arguments.

        :param name: the axis type name, supporting `"ms"`, `"FEs"`,
            `"plainF"`, `"scaledF"`, and `"normalizedF"`
        :param chosen_min: the chosen minimum
        :param chosen_max: the chosen maximum
        :param use_data_min: should the data minimum be used
        :param use_data_max: should the data maximum be used
        :param log_scale: the log scale indicator
        :param log_base: the log base
        :return: the `AxisRanger`
        """
        if not isinstance(name, str):
            raise type_error(name, "axis name", str)

        __log: bool = False
        __min: float | None = None
        __max: float | None = None
        __data_min: bool = chosen_min is None
        __data_max: bool = chosen_max is None

        if name in (TIME_UNIT_MILLIS, KEY_LAST_IMPROVEMENT_TIME_MILLIS,
                    KEY_TOTAL_TIME_MILLIS, KEY_ERT_TIME_MILLIS):
            if chosen_min is not None:
                if (chosen_min < 0) or (not isfinite(chosen_min)):
                    raise ValueError("chosen_min must be >= 0 for axis "
                                     f"type {name}, but is {chosen_min}.")
                __log = (chosen_min > 0)
                if log_scale is not None:
                    if log_scale and (not __log):
                        raise ValueError(f"Cannot set log_scale={log_scale} "
                                         f"and chosen_min={chosen_min} for "
                                         f"axis type {name}.")
                    __log = log_scale
            elif log_scale is None:
                __log = True
            else:
                __log = log_scale

            __min = (1 if __log else 0) if chosen_min is None else chosen_min

            if use_data_max is not None:
                __data_max = use_data_max

            __data_min = False if use_data_min is None else use_data_min

            if chosen_max is not None:
                __max = chosen_max

            return AxisRanger(__min, __max, __data_min, __data_max,
                              __log, log_base if __log else None)

        if name in (TIME_UNIT_FES, KEY_LAST_IMPROVEMENT_FE, KEY_TOTAL_FES,
                    KEY_ERT_FES):
            if chosen_min is None:
                __min = 1
            else:
                if (chosen_min < 1) or (not isfinite(chosen_min)):
                    raise ValueError("chosen_min must be >= 1 for axis "
                                     f"type {name}, but is {chosen_min}.")
                __min = chosen_min
            __log = True if (log_scale is None) else log_scale

            if use_data_max is not None:
                __data_max = use_data_max

            __data_min = False if use_data_min is None else use_data_min

            return AxisRanger(__min, chosen_max, __data_min, __data_max,
                              __log, log_base if __log else None)

        if name == F_NAME_RAW:
            if use_data_max is not None:
                __data_max = use_data_max
            if use_data_min is not None:
                __data_min = use_data_min
            if log_scale is not None:
                __log = log_scale
            return AxisRanger(chosen_min, chosen_max, __data_min, __data_max,
                              __log, log_base if __log else None)

        if name == F_NAME_SCALED:
            __min = 1
        elif name == F_NAME_NORMALIZED:
            if (log_scale is None) or (not log_scale):
                __min = 0
        elif name == "ecdf":
            if (log_scale is None) or (not log_scale):
                __min = 0
            __max = 1
        else:
            raise ValueError(f"Axis type {name!r} is unknown.")

        if chosen_min is not None:
            __min = chosen_min

        if log_scale is not None:
            __log = log_scale
        if use_data_max is not None:
            __data_max = use_data_max
        if use_data_min is not None:
            __data_min = use_data_min

        return AxisRanger(__min, chosen_max, __data_min, __data_max,
                          __log, log_base if __log else None)

    @staticmethod
    def for_axis_func(chosen_min: float | None = None,
                      chosen_max: float | None = None,
                      use_data_min: bool | None = None,
                      use_data_max: bool | None = None,
                      log_scale: bool | None = None,
                      log_base: float | None = None) -> Callable:
        """
        Generate a function that provides the default per-axis ranger.

        :param chosen_min: the chosen minimum
        :param chosen_max: the chosen maximum
        :param use_data_min: should the data minimum be used
        :param use_data_max: should the data maximum be used
        :param log_scale: the log scale indicator
        :param log_base: the log base
        :return: a function in the shape of :meth:`for_axis` with the
            provided defaults
        """
        def __func(name: str,
                   cmi=chosen_min,
                   cma=chosen_max,
                   udmi=use_data_min,
                   udma=use_data_max,
                   ls=log_scale,
                   lb=log_base) -> AxisRanger:
            return AxisRanger.for_axis(name, cmi, cma, udmi, udma, ls, lb)

        return __func



"""
An implementation of NSGA-II, a famous multi-objective optimization method.

Here we implement the famous NSGA-II algorithm by Deb, Agrawal, Pratap, and
Meyarivan.

We apply the following modifications to the computation of the crowding
distance:
First, we normalize each dimension based on its range in the population. This
may make sense because some objective functions may have huge ranges while
others have very small ranges and would then basically be ignored in the
crowding distance.
Also, we do not assign infinite crowding distances to border elements of
collapsed dimensions. In other words, if all elements have the same value of
one objective function, none of them would receive infinite crowding distance
for this objective. This also makes sense because in this case, sorting would
be rather random.

When selecting parents, we apply tournament selection with replacement.

1. Kalyanmoy Deb, Samir Agrawal, Amrit Pratap, and T. Meyarivan. A Fast
   Elitist Non-Dominated Sorting Genetic Algorithm for Multi-Objective
   Optimization: NSGA-II. In Proceedings of the International Conference
   on Parallel Problem Solving from Nature (PPSN VI), September 18-20, 2000,
   Paris, France, pages 849-858. Lecture Notes in Computer Science,
   volume 1917. ISBN 978-3-540-41056-0. Berlin, Heidelberg: Springer.
   https://doi.org/10.1007/3-540-45356-3_83.
   Also: KanGAL Report Number 200001.
   http://repository.ias.ac.in/83498/1/2-a.pdf.

"""
from math import inf
from typing import Any, Callable, Final, cast

import numpy as np
from numpy.random import Generator

from moptipy.api.algorithm import Algorithm2
from moptipy.api.mo_algorithm import MOAlgorithm
from moptipy.api.mo_archive import MORecord
from moptipy.api.mo_process import MOProcess
from moptipy.api.operators import Op0, Op1, Op2
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.strings import num_to_str_for_name
from moptipy.utils.types import type_error


class _NSGA2Record(MORecord):
    """The NSGA-II specific internal record structure."""

    def __init__(self, x: Any, fs: np.ndarray) -> None:
        """
        Create a multi-objective record.

        :param x: the point in the search space
        :param fs: the vector of objective values
        """
        super().__init__(x, fs)
        #: the list of dominated solutions
        self.dominates: Final[list[_NSGA2Record]] = []
        #: the number of solutions this one is dominated by
        self.n_dominated_by: int = 0
        #: the pareto front index
        self.front: int = -1
        #: the crowding distance
        self.crowding_distance: float = 0.0

    def __lt__(self, other):
        """
        Compare according to domination front and crowding distance.

        :param other: the other element to compare with
        :returns: `True` if an only if this solution has either a smaller
            Pareto front index (:attr:`front`) or the same front index
            and a larger crowding distance (:attr:`crowding_distance`)
        """
        fs: Final[int] = self.front
        fo: Final[int] = other.front
        return (fs < fo) or ((fs == fo) and (
            self.crowding_distance > other.crowding_distance))


def _non_dominated_sorting(
        pop: list[_NSGA2Record],
        needed_at_least: int,
        domination: Callable[[np.ndarray, np.ndarray], int]) -> int:
    """
    Perform the fast non-dominated sorting.

    :param pop: the population
    :param needed_at_least: the number of elements we actually need
    :param domination: the domination relationship
    :returns: the end of the last relevant frontier, with
        `needed_at_least <= end`
    """
    lp: Final[int] = len(pop)

    # clear the domination record
    for rec in pop:
        rec.n_dominated_by = 0

    # compute the domination
    for i in range(lp - 1):
        a = pop[i]
        # we only check each pair of solutions once
        for j in range(i + 1, lp):
            b = pop[j]
            d = domination(a.fs, b.fs)
            if d == -1:  # if a dominates b
                a.dominates.append(b)   # remember that it does
                b.n_dominated_by += 1   # increment b's domination count
            elif d == 1:  # if b dominates a
                b.dominates.append(a)   # remember that it does
                a.n_dominated_by += 1   # increment b's domination count
    # now compute the fronts
    done: int = 0
    front_start: int
    front: int = 0
    while True:  # repeat until all records are done
        front_start = done  # remember end of old front
        front += 1          # step front counter
        for i in range(done, lp):  # process all records not yet in a front
            a = pop[i]  # pick record at index i
            if a.n_dominated_by == 0:  # if it belongs to the current front
                a.front = front  # set its front counter
                pop[i], pop[done] = pop[done], a  # swap it to the front end
                done += 1  # increment length of done records
        if done >= needed_at_least:  # are we done with sorting?
            break  # yes -> break loop and return
        # the new front has been built, it ranges from [old_done, done)
        for j in range(front_start, done):
            a = pop[j]
            for b in a.dominates:
                b.n_dominated_by -= 1
            a.dominates.clear()

    # clear the remaining domination records
    for i in range(front_start, len(pop)):
        pop[i].dominates.clear()

    # return the number of elements sorted
    return done


def _crowding_distances(pop: list[_NSGA2Record]) -> None:
    """
    Compute the crowding distances.

    This method works as in the original NSGA-II, but it normalizes each
    dimension based on their ranges in the population. This may make sense
    because some objective functions may have huge ranges while others have
    very small ranges and would then basically be ignored in the crowding
    distance.

    Also, we do not assign infinite crowding distances to border elements of
    collapsed dimensions. In other words, if all elements have the same value
    of one objective function, none of them would receive infinite crowding
    distance for this objective. This also makes sense because in this case,
    sorting would be rather random.

    :param pop: the population
    """
    for rec in pop:  # set all crowding distances to 0
        rec.crowding_distance = 0.0

    dimensions: Final[int] = len(pop[0].fs)
    for dim in range(dimensions):
        pop.sort(key=cast(Callable[[_NSGA2Record], Any],
                          lambda r, d=dim: r.fs[d]))  # sort by dimension
        first = pop[0]  # get the record with the smallest value in the dim
        last = pop[-1]  # get the record with the largest value in the dim
        a = first.fs[dim]  # get smallest value of dimension
        b = last.fs[dim]   # get largest value of dimension
        if a >= b:    # if smallest >= largest, all are the same!
            continue  # we can ignore the dimension if it has collapsed
        first.crowding_distance = inf  # crowding dist of smallest = inf
        last.crowding_distance = inf   # crowding dist of largest = inf
        div = b - a   # get divisor for normalization
        if div <= 0:  # if the range collapses due to numerical imprecision...
            continue  # then we can also skip this dimension
        # ok, we do have non-zero range
        nex = pop[1]  # for each element, we need those to the left and right
        for i in range(2, len(pop)):
            rec = nex     # first: rec = pop[1], second: pop[2], ....
            nex = pop[i]  # first: nex = pop[2], second: pop[3], ....
            rec.crowding_distance += (nex.fs[dim] - first.fs[dim]) / div
            first = rec   # current rec now = rec to the left next


class NSGA2(Algorithm2, MOAlgorithm):
    """The NSGA-II algorithm by Deb, Agrawal, Pratap, and Meyarivan."""

    def __init__(self, op0: Op0, op1: Op1, op2: Op2,
                 pop_size: int, cr: float) -> None:
        """
        Create the NSGA-II algorithm.

        :param op0: the nullary search operator
        :param op1: the unary search operator
        :param op2: the binary search operator
        :param pop_size: the population size
        :param cr: the crossover rate
        """
        super().__init__(
            f"nsga2_{pop_size}_{num_to_str_for_name(cr)}", op0, op1, op2)
        if not isinstance(pop_size, int):
            raise type_error(pop_size, "pop_size", int)
        if pop_size < 3:
            raise ValueError(f"pop_size={pop_size} < 3.")
        #: the population size
        self.pop_size: Final[int] = pop_size
        if not isinstance(cr, float):
            raise type_error(cr, "cr", float)
        if not 0 < cr < 1:
            raise ValueError(f"cr={cr}, but we need 0<cr<1.")
        #: the crossover rate
        self.cr: Final[float] = cr

    def solve_mo(self, process: MOProcess) -> None:
        """
        Apply the MO-RLS to an optimization problem.

        :param process: the black-box process object
        """
        # initialize fast calls and local constants
        create_x: Final[Callable[[], Any]] = process.create
        create_f: Final[Callable[[], np.ndarray]] = process.f_create
        domination: Final[Callable[[np.ndarray, np.ndarray], int]] = \
            process.f_dominates
        should_terminate: Final[Callable[[], bool]] = process.should_terminate
        random: Final[Generator] = process.get_random()
        op0: Final[Callable[[Generator, Any], None]] = self.op0.op0
        op1: Final[Callable[[Generator, Any, Any], None]] = self.op1.op1
        op2: Final[Callable[[Generator, Any, Any, Any], None]] = self.op2.op2
        evaluate: Final[Callable[[Any, np.ndarray], Any]] = process.f_evaluate
        ri: Final[Callable[[int], int]] = random.integers
        rd: Final[Callable[[], float]] = random.uniform
        cr: Final[float] = self.cr

        # create first population
        pop_size: Final[int] = self.pop_size
        pop_size_2: Final[int] = pop_size + pop_size
        pop: list[_NSGA2Record] = []

        # create the population
        for _ in range(pop_size):
            if should_terminate():
                return
            rec: _NSGA2Record = _NSGA2Record(create_x(), create_f())
            op0(random, rec.x)
            evaluate(rec.x, rec.fs)
            pop.append(rec)
        _non_dominated_sorting(pop, pop_size, domination)

        # create offspring records (initially empty)
        for _ in range(pop_size):
            pop.append(_NSGA2Record(create_x(), create_f()))

        leave: bool = False
        while True:  # the main loop

            # fill offspring population
            for ofs_idx in range(pop_size, pop_size_2):
                # check if we should leave before evaluation
                if should_terminate():
                    leave = True  # oh, we need to quit
                    break  # break inner loop

                ofs: _NSGA2Record = pop[ofs_idx]  # offspring to overwrite
                # binary tournament with replacement for first parent
                p1: _NSGA2Record = pop[ri(pop_size)]
                palt: _NSGA2Record = pop[ri(pop_size)]
                if palt < p1:
                    p1 = palt

                if rd() >= cr:  # mutation: only 1 parent needed
                    op1(random, ofs.x, p1.x)  # do mutation
                else:  # crossover: need a second parent
                    # binary tournament with replacement for second paren
                    # (who must be different from first parent)
                    p2: _NSGA2Record = p1
                    while p2 is p1:
                        p2 = pop[ri(pop_size)]
                    palt = p1
                    while palt is p1:
                        palt = pop[ri(pop_size)]
                    if palt < p2:
                        p2 = palt
                    # got two parents, now do crossover
                    op2(random, ofs.x, p1.x, p2.x)
                evaluate(ofs.x, ofs.fs)  # otherwise: evaluate

            if leave:  # should we leave?
                break  # yes! break main loop

            # Perform non-dominated sorting and get at least pop_size
            # elements in different fronts.
            # "start" marks the begin of the last front that was created.
            # "end" marks the total number of elements sorted.
            # It holds that "start < pop_size <= end".
            end = _non_dominated_sorting(pop, pop_size, domination)
            # We only perform the crowding distance computation on the first
            # "end" elements in the population.
            # We take this slice of the population (or the whole population if
            # all elements are non-domination) and compute the crowding
            # distance.
            candidates = pop[:end] if end < pop_size_2 else pop
            _crowding_distances(candidates)
            # Careful: crowding distance computation will shuffle the list
            # candidates, because it sorts based on dimensions. Therefore, we
            # need to re-sort the whole slice.
            candidates.sort()
            if candidates is not pop:  # write back if need be
                pop[:end] = candidates

        # At the end we flush population to the process archive.
        for rec in pop:
            if rec.front != -1:
                process.check_in(rec.x, rec.fs)

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of the algorithm to a logger.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        logger.key_value("pop_size", self.pop_size)
        logger.key_value("cr", self.cr)

    def initialize(self) -> None:
        """Initialize the algorithm."""
        Algorithm2.initialize(self)



"""Some internal helper functions and base classes."""

from dataclasses import dataclass
from math import inf
from typing import Final

from moptipy.utils.nputils import rand_seed_check
from moptipy.utils.strings import sanitize_name
from moptipy.utils.types import type_error

#: The key for the total number of runs.
KEY_N: Final[str] = "n"

#: The unit of the time axis if time is measured in milliseconds.
TIME_UNIT_MILLIS: Final[str] = "ms"
#: The unit of the time axis of time is measured in FEs
TIME_UNIT_FES: Final[str] = "FEs"

#: The name of the raw objective values data.
F_NAME_RAW: Final[str] = "plainF"
#: The name of the scaled objective values data.
F_NAME_SCALED: Final[str] = "scaledF"
#: The name of the normalized objective values data.
F_NAME_NORMALIZED: Final[str] = "normalizedF"


def check_time_unit(time_unit: str) -> str:
    """
    Check that the time unit is OK.

    :param time_unit: the time unit
    :return: the time unit string
    """
    if not isinstance(time_unit, str):
        raise type_error(time_unit, "time_unit", str)
    if time_unit in (TIME_UNIT_FES, TIME_UNIT_MILLIS):
        return time_unit
    raise ValueError(
        f"Invalid time unit {time_unit!r}, only {TIME_UNIT_FES!r} "
        f"and {TIME_UNIT_MILLIS!r} are permitted.")


def check_f_name(f_name: str) -> str:
    """
    Check whether an objective value name is valid.

    :param f_name: the name of the objective function dimension
    :return: the name of the objective function dimension
    """
    if not isinstance(f_name, str):
        raise type_error(f_name, "f_name", str)
    if f_name in (F_NAME_RAW, F_NAME_SCALED, F_NAME_NORMALIZED):
        return f_name
    raise ValueError(
        f"Invalid f name {f_name!r}, only {F_NAME_RAW}, "
        f"{F_NAME_SCALED}, and {F_NAME_NORMALIZED} are permitted.")


@dataclass(frozen=True, init=False, order=True)
class PerRunData:
    """
    An immutable record of information over a single run.

    >>> p = PerRunData("a", "i", 234)
    >>> print(p.instance)
    i
    >>> print(p.algorithm)
    a
    >>> print(p.rand_seed)
    234
    """

    #: The algorithm that was applied.
    algorithm: str

    #: The problem instance that was solved.
    instance: str

    #: The seed of the random number generator.
    rand_seed: int

    def __init__(self, algorithm: str, instance: str, rand_seed: int):
        """
        Create a per-run data record.

        :param str algorithm: the algorithm name
        :param str instance: the instance name
        :param int rand_seed: the random seed
        """
        if not isinstance(algorithm, str):
            raise type_error(algorithm, "algorithm", str)
        if algorithm != sanitize_name(algorithm):
            raise ValueError(f"Invalid algorithm name {algorithm!r}.")
        object.__setattr__(self, "algorithm", algorithm)

        if not isinstance(instance, str):
            raise type_error(instance, "instance", str)
        if instance != sanitize_name(instance):
            raise ValueError(f"Invalid instance name {instance!r}.")
        object.__setattr__(self, "instance", instance)
        object.__setattr__(self, "rand_seed", rand_seed_check(rand_seed))


@dataclass(frozen=True, init=False, order=True)
class MultiRunData:
    """
    A class that represents statistics over a set of runs.

    If one algorithm*instance is used, then `algorithm` and `instance` are
    defined. Otherwise, only the parameter which is the same over all recorded
    runs is defined.

    >>> p = MultiRunData("a", "i", 3)
    >>> print(p.instance)
    i
    >>> print(p.algorithm)
    a
    >>> print(p.n)
    3
    """

    #: The algorithm that was applied, if the same over all runs.
    algorithm: str | None
    #: The problem instance that was solved, if the same over all runs.
    instance: str | None
    #: The number of runs over which the statistic information is computed.
    n: int

    def __init__(self,
                 algorithm: str | None,
                 instance: str | None,
                 n: int):
        """
        Create the dataset of an experiment-setup combination.

        :param algorithm: the algorithm name, if all runs are with the same
            algorithm
        :param instance: the instance name, if all runs are on the same
            instance
        :param n: the total number of runs
        """
        if (algorithm is not None) and \
                (algorithm != sanitize_name(algorithm)):
            raise ValueError(f"Invalid algorithm {algorithm!r}.")
        object.__setattr__(self, "algorithm", algorithm)

        if (instance is not None) and (instance != sanitize_name(instance)):
            raise ValueError(f"Invalid instance {instance!r}.")
        object.__setattr__(self, "instance", instance)

        if not isinstance(n, int):
            raise type_error(n, "n", int)
        if n <= 0:
            raise ValueError(f"n must be > 0, but is {n}.")
        object.__setattr__(self, "n", n)


@dataclass(frozen=True, init=False, order=True)
class MultiRun2DData(MultiRunData):
    """
    A multi-run data based on one time and one objective dimension.

    >>> p = MultiRun2DData("a", "i", 3, TIME_UNIT_FES, F_NAME_SCALED)
    >>> print(p.instance)
    i
    >>> print(p.algorithm)
    a
    >>> print(p.n)
    3
    >>> print(p.time_unit)
    FEs
    >>> print(p.f_name)
    scaledF
    """

    #: The unit of the time axis.
    time_unit: str
    #: the name of the objective value axis.
    f_name: str

    def __init__(self,
                 algorithm: str | None,
                 instance: str | None,
                 n: int,
                 time_unit: str,
                 f_name: str):
        """
        Create multi-run data based on one time and one objective dimension.

        :param algorithm: the algorithm name, if all runs are with the same
            algorithm
        :param instance: the instance name, if all runs are on the same
            instance
        :param n: the total number of runs
        :param time_unit: the time unit
        :param f_name: the objective dimension name
        """
        super().__init__(algorithm, instance, n)

        object.__setattr__(self, "time_unit", check_time_unit(time_unit))
        object.__setattr__(self, "f_name", check_f_name(f_name))


def get_instance(obj: PerRunData | MultiRunData) -> str | None:
    """
    Get the instance of a given object.

    :param obj: the object
    :return: the instance string, or `None` if no instance is specified

    >>> p1 = MultiRunData("a", "i1", 3)
    >>> print(get_instance(p1))
    i1
    >>> p2 = PerRunData("a", "i2", 31)
    >>> print(get_instance(p2))
    i2
    """
    return obj.instance


def get_algorithm(obj: PerRunData | MultiRunData) -> str | None:
    """
    Get the algorithm of a given object.

    :param obj: the object
    :return: the algorithm string, or `None` if no algorithm is specified

    >>> p1 = MultiRunData("a1", "i1", 3)
    >>> print(get_algorithm(p1))
    a1
    >>> p2 = PerRunData("a2", "i2", 31)
    >>> print(get_algorithm(p2))
    a2
    """
    return obj.algorithm


def sort_key(obj: PerRunData | MultiRunData) -> \
        tuple[str, str, str, int, int, str, str, float]:
    """
    Get a default sort key for the given object.

    The sort key is a tuple with well-defined field elements that should
    allow for a default and consistent sorting over many different elements of
    the experiment evaluation data API. Sorting should work also for lists
    containing elements of different classes.

    :param obj: the object
    :return: the sort key

    >>> p1 = MultiRunData("a1", "i1", 3)
    >>> p2 = PerRunData("a2", "i2", 31)
    >>> print(sort_key(p1) < sort_key(p2))
    True
    >>> print(sort_key(p1) >= sort_key(p2))
    False
    >>> p3 = MultiRun2DData("a", "i", 3, TIME_UNIT_FES, F_NAME_SCALED)
    >>> print(sort_key(p3) < sort_key(p1))
    True
    >>> print(sort_key(p3) >= sort_key(p1))
    False
    """
    goal_f = getattr(obj, "goal_f") if hasattr(obj, "goal_f") else None
    if goal_f is None:
        goal_f = inf

    return obj.__class__.__name__, \
        "" if obj.algorithm is None else obj.algorithm, \
        "" if obj.instance is None else obj.instance, \
        obj.n if isinstance(obj, MultiRunData) else 0, \
        obj.rand_seed if isinstance(obj, PerRunData) else 0, \
        obj.time_unit if isinstance(obj, MultiRun2DData) else "", \
        obj.f_name if isinstance(obj, MultiRun2DData) else "", \
        goal_f



"""
Ackley's Function.

Ackley's function is a continuous, multimodal, differentiable, non-separable,
and scalable benchmark function for continuous optimization.

1. David H. Ackley. *A Connectionist Machine for Genetic Hillclimbing.* 1987.
   Volume 28 of The Kluwer International Series in Engineering and Computer
   Science. Norwell, MA, USA: Kluwer Academic Publisher.
   ISBN: 978-1-4612-9192. doi:10.1007/978-1-4613-199
"""

import numba  # type: ignore
from numpy import cos, e, exp, mean, ndarray, pi, sqrt

from moptipy.api.objective import Objective


@numba.njit(nogil=True, cache=True)
def ackley(x: ndarray) -> float:
    """
    Compute Ackley's function.

    :param x: the np array
    :return: the result of Ackley's function

    >>> from numpy import array
    >>> print(abs(ackley(array([0.0, 0.0, 0.0]))) < 1e-15)
    True
    >>> print(ackley(array([2.0, 2.0])))
    6.593599079287213
    >>> print(ackley(array([-3.0, 2.0])))
    7.9889108105187
    """
    res = 20.0 + e + (-20.0 * exp(-0.2 * sqrt(
        mean(x ** 2)))) - exp(mean(cos(2.0 * pi * x)))
    return 0.0 if res <= 0.0 else float(res)


class Ackley(Objective):
    """Ackley's function."""

    def __init__(self) -> None:
        """Initialize Ackley's function."""
        self.evaluate = ackley  # type: ignore

    def lower_bound(self) -> float:
        """
        Get the lower bound of the sphere problem.

        :return: 0

        >>> print(Ackley().lower_bound())
        0.0
        """
        return 0.0

    def __str__(self) -> str:
        """
        Get the name of Ackley's function.

        :return: `ackley`
        :retval "ackley": always

        >>> print(Ackley())
        ackley
        """
        return "ackley"



"""Shared constants and functions for dealing with logs."""

from typing import Final

#: the file suffix to be used for log files
FILE_SUFFIX: Final[str] = ".txt"

#: the key for the exception type
KEY_EXCEPTION_TYPE: Final[str] = "exceptionType"
#: the key for the exception value
KEY_EXCEPTION_VALUE: Final[str] = "exceptionValue"
#: the key for the exception stack trace
KEY_EXCEPTION_STACK_TRACE: Final[str] = "exceptionStackTrace"

#: the key for algorithms
KEY_ALGORITHM: Final[str] = "algorithm"
#: the key for the instance
KEY_INSTANCE: Final[str] = "instance"

#: the common name prefix of all error sections
ERROR_SECTION_PREFIX: Final[str] = "ERROR_"
#: the section indicating an error during the algorithm run
SECTION_ERROR_IN_RUN: Final[str] = f"{ERROR_SECTION_PREFIX}IN_RUN"
#: the section indicating an error that occurred in the context of the
#: process. this may be an error in the algorithm or, more likely, in the
#: processing of the result.
SECTION_ERROR_IN_CONTEXT: Final[str] = f"{ERROR_SECTION_PREFIX}IN_CONTEXT"
#: the section indicating an invalid candidate solution
SECTION_ERROR_INVALID_Y: Final[str] = f"{ERROR_SECTION_PREFIX}INVALID_Y"
#: the section indicating a mismatch of the computed and registered best f
SECTION_ERROR_BEST_F: Final[str] = f"{ERROR_SECTION_PREFIX}BEST_F_MISMATCH"
#: the section indicating an invalid point in the search space
SECTION_ERROR_INVALID_X: Final[str] = f"{ERROR_SECTION_PREFIX}INVALID_X"
#: the section indicating that the time measurement has an error
SECTION_ERROR_TIMING: Final[str] = f"{ERROR_SECTION_PREFIX}TIMING"
#: the section indicating an error caught during log writing
SECTION_ERROR_IN_LOG: Final[str] = f"{ERROR_SECTION_PREFIX}IN_LOG"

#: the progress csv section
SECTION_PROGRESS: Final[str] = "PROGRESS"
#: the FEs column for the progress CSV
PROGRESS_FES: Final[str] = "fes"
#: the time millis column for the progress CSV
PROGRESS_TIME_MILLIS: Final[str] = "timeMS"
#: the current objective value column for the progress CSV
PROGRESS_CURRENT_F: Final[str] = "f"

#: the end state
SECTION_FINAL_STATE: Final[str] = "STATE"
#: the total number of consumed FEs
KEY_TOTAL_FES: Final[str] = "totalFEs"
#: the total number of consumed milliseconds
KEY_TOTAL_TIME_MILLIS: Final[str] = "totalTimeMillis"
#: the best encountered objective value
KEY_BEST_F: Final[str] = "bestF"
#: the vector of objective values of the best solution
KEY_BEST_FS: Final[str] = "bestFs"
#: the number of archived non-dominated solutions
KEY_ARCHIVE_SIZE: Final[str] = "archiveSize"
#: the FE when the best objective value was reached
KEY_LAST_IMPROVEMENT_FE: Final[str] = "lastImprovementFE"
#: the time in milliseconds when the best objective value was reached
KEY_LAST_IMPROVEMENT_TIME_MILLIS: Final[str] = "lastImprovementTimeMillis"

#: the archive section prefix
PREFIX_SECTION_ARCHIVE: Final[str] = "ARCHIVE_"
#: the archive quality section suffix
SECTION_ARCHIVE_QUALITY: Final[str] = f"{PREFIX_SECTION_ARCHIVE}QUALITIES"
#: the scalarized objective value column of the archive if without numeric
#: suffix, the result of the ith objective function if with numeric suffix
KEY_ARCHIVE_F: Final[str] = PROGRESS_CURRENT_F

#: the archive X section suffix
SUFFIX_SECTION_ARCHIVE_X: Final[str] = "_X"
#: the archive Y section suffix
SUFFIX_SECTION_ARCHIVE_Y: Final[str] = "_Y"

#: the setup section
SECTION_SETUP: Final[str] = "SETUP"
#: the default log key for names of objects
KEY_NAME: Final[str] = "name"
#: the class of an object
KEY_CLASS: Final[str] = "class"
#: the inner class of an object
KEY_INNER_CLASS: Final[str] = "innerClass"
#: the default log key for the lower bound of objective function values
KEY_F_LOWER_BOUND: Final[str] = "lowerBound"
#: the default log key for the upper bound of objective function values
KEY_F_UPPER_BOUND: Final[str] = "upperBound"
#: the maximum FEs of a black-box process
KEY_MAX_FES: Final[str] = "maxFEs"
#: the maximum runtime in milliseconds of a black-box process
KEY_MAX_TIME_MILLIS: Final[str] = "maxTimeMillis"
#: the goal objective value of a black-box process
KEY_GOAL_F: Final[str] = "goalF"
#: the random seed
KEY_RAND_SEED: Final[str] = "randSeed"
#: the random generator type
KEY_RAND_GENERATOR_TYPE: Final[str] = "randGenType"
#: the type of the bit generator used by the random generator
KEY_RAND_BIT_GENERATOR_TYPE: Final[str] = "randBitGenType"
#: the number of decision variables
KEY_SPACE_NUM_VARS: Final[str] = "nvars"
#: the maximum archive size (after pruning)
KEY_ARCHIVE_MAX_SIZE: Final[str] = "archiveMaxSize"
#: the pruning limit of the archive size
KEY_ARCHIVE_PRUNE_LIMIT: Final[str] = "archivePruneLimit"

#: the scope of the process parameters
SCOPE_PROCESS: Final[str] = "p"
#: the scope of the solution space
SCOPE_SOLUTION_SPACE: Final[str] = "y"
#: the scope of the search space
SCOPE_SEARCH_SPACE: Final[str] = "x"
#: the scope of the objective function
SCOPE_OBJECTIVE_FUNCTION: Final[str] = "f"
#: the scope of the encoding
SCOPE_ENCODING: Final[str] = "g"
#: the scope of the optimization algorithm
SCOPE_ALGORITHM: Final[str] = "a"
#: the scope of the multi-objective archive pruner
SCOPE_PRUNER: Final[str] = "ap"
#: the scope of the nullary search operator
SCOPE_OP0: Final[str] = "op0"
#: the scope of the unary search operator
SCOPE_OP1: Final[str] = "op1"
#: the scope of the binary search operator
SCOPE_OP2: Final[str] = "op2"

#: the automatically generated system info section
SECTION_SYS_INFO: Final[str] = "SYS_INFO"
#: information about the current session
SCOPE_SESSION: Final[str] = "session"
#: the time when the session was started
KEY_SESSION_START: Final[str] = "start"
#: the node on which the session is running
KEY_NODE_NAME: Final[str] = "node"
#: the affinity of the process to logical CPUs
KEY_CPU_AFFINITY: Final[str] = "cpuAffinity"
#: the pid of the process
KEY_PROCESS_ID: Final[str] = "procesId"
#: the ip address of the node on which the session is running
KEY_NODE_IP: Final[str] = "ipAddress"
#: the versions scope in the sys-info section
SCOPE_VERSIONS: Final[str] = "version"
#: the hardware scope in the sys-info section
SCOPE_HARDWARE: Final[str] = "hardware"
#: the number of physical CPUs
KEY_HW_N_PHYSICAL_CPUS: Final[str] = "nPhysicalCpus"
#: the number of logical CPUs
KEY_HW_N_LOGICAL_CPUS: Final[str] = "nLogicalCpus"
#: the clock speed of the CPUs
KEY_HW_CPU_MHZ: Final[str] = "cpuMhz"
#: the key for the byte order
KEY_HW_BYTE_ORDER: Final[str] = "byteOrder"
#: the key for the machine
KEY_HW_MACHINE: Final[str] = "machine"
#: the key for the cpu name
KEY_HW_CPU_NAME: Final[str] = "cpu"
#: the key for the memory size
KEY_HW_MEM_SIZE: Final[str] = "memSize"
#: the operating system in the sys-info section
SCOPE_OS: Final[str] = "os"
#: the operating system name
KEY_OS_NAME: Final[str] = "name"
#: the operating system version
KEY_OS_VERSION: Final[str] = "version"
#: the operating system release
KEY_OS_RELEASE: Final[str] = "release"
#: the python scope in the sys-info section
SCOPE_PYTHON: Final[str] = "python"
#: the python version
KEY_PYTHON_VERSION: Final[str] = "version"
#: the python implementation
KEY_PYTHON_IMPLEMENTATION: Final[str] = "implementation"

#: the resulting point in the solution space
SECTION_RESULT_Y: Final[str] = "RESULT_Y"
#: the resulting point in the search space
SECTION_RESULT_X: Final[str] = "RESULT_X"

#: all the section titles combined in one set
# noinspection PyDefaultArgument
# pylint: disable=[C3002]
_ALL_SECTIONS: Final[set[str]] = (lambda g=globals(): {  # type: ignore # noqa
    g[k] for k in g if k.startswith("SECTION_")})()  # type: ignore



"""
Direct fitness assignment uses the objective values directly as fitness.

Together with fitness proportionate selection (:class:`~moptipy.algorithms.\
modules.selections.fitness_proportionate_sus.FitnessProportionateSUS`), it
turns an EA (:class:`~moptipy.algorithms.so.general_ea.GeneralEA`) to a
minimization variant of the traditional Genetic Algorithms.

1. John Henry Holland. *Adaptation in Natural and Artificial Systems: An
   Introductory Analysis with Applications to Biology, Control, and Artificial
   Intelligence.* Ann Arbor, MI, USA: University of Michigan Press. 1975.
   ISBN: 0-472-08460-7
2. David Edward Goldberg. *Genetic Algorithms in Search, Optimization, and
   Machine Learning.* Boston, MA, USA: Addison-Wesley Longman Publishing Co.,
   Inc. 1989. ISBN: 0-201-15767-5
"""


from numpy.random import Generator

from moptipy.algorithms.so.fitness import Fitness, FRecord


# start book
class Direct(Fitness):
    """
    Objective values are used as fitness directly.

    This is the most primitive and basic fitness assignment strategy.
    It does not give preference to new solutions over old solutions if both
    have the same objective value. In this, it is different from
    :class:`~moptipy.algorithms.so.fitnesses.rank_and_iteration\
.RankAndIteration`, which is the default fitness assignment strategy.
    It is therefore also not compatible to our basic (mu+lambda) Evolutionary
    Algorithm implementation (:class:`~moptipy.algorithms.so.ea.EA`).
    """

    def assign_fitness(self, p: list[FRecord], random: Generator) -> None:
        """
        Assign the objective value as fitness.

        :param p: the list of records
        :param random: ignored
        """
        for rec in p:  # iterate over all records in the population
            rec.fitness = rec.f  # store objective value f as fitness
# end book

    def __str__(self):
        """
        Get the name of this direct fitness assignment.

        :return: the name of this fitness assignment strategy
        :retval "direct": always
        """
        return "direct"



"""The numeric format definitions."""

from math import inf, isfinite
from typing import Callable, Final, Iterable, cast

from moptipy.utils.formatted_string import FormattedStr
from moptipy.utils.lang import Lang
from moptipy.utils.math import try_int
from moptipy.utils.types import type_error


def default_get_int_renderer() -> Callable[[int], str]:
    """
    Get the default integer renderer.

    :returns: the default integer renderer, which uses the integer rendering
        of the currently active language setting.

    >>> from moptipy.utils.lang import EN, ZH
    >>> EN.set_current()
    >>> f = default_get_int_renderer()
    >>> f(1_000_000)
    "1'000'000"
    >>> ZH.set_current()
    >>> f = default_get_int_renderer()
    >>> f(1_000_000)
    "100'0000"
    """
    return cast(Callable[[int], str], Lang.current().format_int)


def default_get_float_format(
        min_finite: int | float = 0,
        max_finite: int | float = 0,
        max_frac_len: int = 2,
        min_non_zero_abs: int | float = inf,
        int_to_float_threshold: int | float = 10_000_000_000) -> str:
    """
    Get the default float format for numbers in the given range.

    :param min_finite: the minimum finite value that may need to be formatted
    :param max_finite: the maximum finite value that may need to be formatted
    :param max_frac_len: the length of the longest fractional part of any
        number encountered that can be converted to a string *not* in the "E"
        notation
    :param min_non_zero_abs: the minimum non-zero absolute value; will be
        `inf` if all absolute values are zero
    :param int_to_float_threshold: the threshold above which all integers are
        converted to floating point numbers with the 'E' notation

    >>> default_get_float_format(0, 0, 0)
    '{:.0f}'
    >>> default_get_float_format(0, 1e5, 10)
    '{:.0f}'
    >>> default_get_float_format(-1e7, 1e2, 10)
    '{:.0f}'
    >>> default_get_float_format(0, 0, 1)
    '{:.1f}'
    >>> default_get_float_format(0, 1e3, 11)
    '{:.1f}'
    >>> default_get_float_format(-1e3, 1e2, 11)
    '{:.1f}'
    >>> default_get_float_format(0, 0, 2)
    '{:.2f}'
    >>> default_get_float_format(0, 0, 3)
    '{:.3f}'
    >>> default_get_float_format(0, 0, 4)
    '{:.3f}'
    >>> default_get_float_format(0, 1e11, 4)
    '{:.2e}'
    >>> default_get_float_format(-1, 1, 4, 1e-3)
    '{:.3f}'
    """
    if not isinstance(min_finite, (int, float)):
        raise type_error(min_finite, "min_finite", (int, float))
    if not isinstance(max_finite, (int, float)):
        raise type_error(max_finite, "max_finite", (int, float))
    if not (isfinite(min_finite) and isfinite(max_finite)
            and (min_finite <= max_finite)):
        raise ValueError("invalid min_finite, max_finite pair "
                         f"{min_finite}, {max_finite}.")
    if not isinstance(max_frac_len, int):
        raise type_error(max_frac_len, "frac_len", int)
    if not (0 <= max_frac_len < 100):
        raise ValueError(f"invalid frac_len {max_frac_len}.")
    if not isinstance(int_to_float_threshold, (int, float)):
        raise type_error(
            int_to_float_threshold, "int_to_float_threshold", (int, float))
    if (int_to_float_threshold <= 0) or not (
            isfinite(int_to_float_threshold)
            or (int_to_float_threshold >= inf)):
        raise ValueError(
            f"invalid int_to_float_threshold={int_to_float_threshold}.")
    if not isinstance(min_non_zero_abs, (int, float)):
        raise type_error(min_non_zero_abs, "min_non_zero_abs", (int, float))
    if min_non_zero_abs <= 0:
        raise ValueError(f"invalid min_non_zero_abs={min_non_zero_abs}")

    # are the values in the [-1, 1] range, i.e., possibly just small fractions?
    if (min_finite >= -1) and (max_finite <= 1) and isfinite(min_non_zero_abs):
        if min_non_zero_abs >= 1e-1:
            return "{:.1f}"
        if min_non_zero_abs >= 1e-2:
            return "{:.2f}"
        if min_non_zero_abs >= 1e-3:
            return "{:.3f}"
        if min_non_zero_abs >= 1e-4:
            return "{:.4f}"
        return "{:.3e}"

    # handle numbers that are outside [-1, 1]
    if ((-int_to_float_threshold) <= min_finite) \
            and (max_finite <= int_to_float_threshold):
        if (max_frac_len <= 0) or (min_finite <= -1E4) or (max_finite >= 1E4):
            return "{:.0f}"
        if (max_frac_len <= 1) or (min_finite <= -1E3) or (max_finite >= 1E3):
            return "{:.1f}"
        if (max_frac_len <= 2) or (min_finite <= -1E2) or (max_finite >= 1E2):
            return "{:.2f}"
        return "{:.3f}"
    return "{:.2e}"


class NumberRenderer:
    """
    A format description for a group of numbers.

    With instances of this class, you can convert a sequence of numbers
    to a sequence of strings with uniform, pleasant formatting. The idea
    is that such numbers can be written, e.g., into a column of a table
    and that this column will then have a nice and uniform appearance.
    In other words, we will avoid situations like the following:
    "1234938845, 1e-20, 0.002, 34757773, 1e30, 0.9998837467"
    which looks rather odd. While the numbers may be displayed correctly,
    the formatting of all numbers is different. If we want to present
    numbers that describe related quantities, we rather want them to all
    have the same format. This class here can achieve this in a customizable
    way.
    """

    def __init__(self,
                 int_to_float_threshold: int | float = 10_000_000_000,
                 get_int_renderer: Callable[[], Callable[[int], str]]
                 = default_get_int_renderer,
                 get_float_format: Callable[
                     [int | float, int | float, int,
                      int | float, int | float], str] =
                 default_get_float_format):
        """
        Create the number group format.

        :param int_to_float_threshold: the threshold above which integers are
            converted to floating point numbers in the 'E' notation.
        :param get_int_renderer: the function to be used to get the renderer
            for all integers and integer parts of floats.
        :param get_float_format: the getter for the float format, i.e., a
            callable accepting the range [min, max] of all finite values to be
            rendered, the maximum length of any fractional part, and the
            `int_to_float_threshold` value and that then returns a string with
            the float format definition
        """
        super().__init__()
        while True:
            if not isinstance(int_to_float_threshold, (int, float)):
                raise type_error(int_to_float_threshold,
                                 "int_to_float_threshold", (int, float))
            if (int_to_float_threshold <= 0) or not (
                    isfinite(int_to_float_threshold)
                    or (int_to_float_threshold >= inf)):
                raise ValueError("invalid int_to_float_threshold="
                                 f"{int_to_float_threshold}.")
            if isinstance(int_to_float_threshold, float):
                a = int(int_to_float_threshold)
                if a == int_to_float_threshold:
                    int_to_float_threshold = a
            else:
                break
        #: the absolute threshold above which all integer numbers must be
        #: converted to floats to render them in the 'E' notation
        self.int_to_float_threshold: Final[int | float] \
            = int_to_float_threshold
        if not callable(get_int_renderer):
            raise type_error(get_int_renderer, "int_renderer", call=True)
        #: the function to be used to get the renderer for all integers and
        #: integer parts of floats
        self.get_int_renderer: Final[Callable[[], Callable[[int], str]]] \
            = get_int_renderer
        #: the getter for the float format to be used to represent a range of
        #: values
        self.get_float_format: Final[Callable[
            [int | float, int | float, int,
             int | float, int | float], str]] = get_float_format

    def derive(self,
               int_to_float_threshold: int | float | None = None,
               get_int_renderer: Callable[[], Callable[
                   [int], str]] | None = None,
               get_float_format: Callable[[int | float, int | float, int,
                                           int | float, int | float],
                                          str] | None = None) \
            -> "NumberRenderer":
        """
        Derive a new number group format from this one.

        :param int_to_float_threshold: the int-to-float threshold
        :param get_int_renderer: the integer renderer getter
        :param get_float_format: the float format getter
        :returns: a new number group format that differs from the current
            format only in terms of the non-`None` parameters specified

        >>> d = DEFAULT_NUMBER_RENDERER
        >>> d.derive() is d
        True
        >>> d.int_to_float_threshold
        10000000000
        >>> from moptipy.utils.lang import EN
        >>> EN.set_current()
        >>> d.get_int_renderer()(123456789)
        "123'456'789"
        >>> d.get_float_format(-10, 10, 2)
        '{:.2f}'
        >>> d = d.derive(int_to_float_threshold=22)
        >>> d is DEFAULT_NUMBER_RENDERER
        False
        >>> d.int_to_float_threshold
        22
        >>> d = d.derive(get_int_renderer=lambda: lambda x: "bla")
        >>> d.get_int_renderer()(112)
        'bla'
        """
        # pylint: disable=R0916
        if (((int_to_float_threshold is None)
             or (int_to_float_threshold == self.int_to_float_threshold))
                and ((get_int_renderer is None)
                     or (get_int_renderer is self.get_int_renderer))
                and ((get_float_format is None)
                     or (get_float_format is self.get_float_format))):
            return self
        return NumberRenderer(
            self.int_to_float_threshold if
            int_to_float_threshold is None else int_to_float_threshold,
            self.get_int_renderer if get_int_renderer is None
            else get_int_renderer,
            self.get_float_format if get_float_format is None
            else get_float_format)

    def render(self, source: int | float | None | Iterable[int | float | None],
               none_str: FormattedStr | None = None) \
            -> list[FormattedStr | None]:
        r"""
        Convert a sequence of numbers to text with uniform shape.

        Often, we need to convert a set of numbers to strings as output for a
        table or another representative thext. In such a case, you want to
        present all numbers in the set in the same format.

        Imagine you have the number vector `[1E-4, 1/7, 123456789012345678]`.
        If you simply convert this list to a string directly, what you get is
        `[0.0001, 0.14285714285714285, 123456789012345678]`. Now this looks
        very ugly. First, we have one very big number `123456789012345678`.
        If the numbers stem from an experiment, then we are hardly able to
        obtain any number at a very extreme precision. The 18 digits in
        `123456789012345678` sort of suggest a precision to 18 decimals, since
        the number ends in specific digits (as opposed to `123450000000000000`
        which a reader would naturally preceive as a rounded quantity).
        Additionally, we have the number `0.14285714285714285`, which has a
        very long fractional part, which, too, suggests a very high precision.
        Writing both mentioned numbers next to each other, this suggests as if
        we could present a number as high as 10**18 at a precision of 10**-17.
        And it also looks ugly, because both numbers are not uniformly
        formatted. Instead, our function here renders the number list as
        `['1.00*10^-4^', '1.43*10^-1^', '1.23*10^17^']`. It recognizes that we
        should present numbers as powers of ten and then limits the precision
        to three digits.

        This function is thus intended to produce some sort of uniform format
        with reasonable precision uniformly for a numerical vector, under the
        assumption that all numbers should be presented in the same numerical
        range and quantity.

        :param source: the column data
        :param none_str: the string replacement for `None`
        :returns: a list with the text representation

        >>> from moptipy.utils.lang import EN
        >>> EN.set_current()
        >>> ff = DEFAULT_NUMBER_RENDERER
        >>> ff.render([1.75651, 212, 3234234])
        ['2', '212', "3'234'234"]
        >>> ff.render([1.75651, 22, 34])
        ['1.757', '22.000', '34.000']
        >>> ff.render([1.75651, 122, 34])
        ['1.76', '122.00', '34.00']
        >>> ff.render([1.75651, 122, 3334])
        ['1.8', '122.0', "3'334.0"]
        >>> ff.render([1.5, 212, 3234234])
        ['2', '212', "3'234'234"]
        >>> ff.render([1.5, 2e12, 3234234])
        ['1.50e0', '2.00e12', '3.23e6']
        >>> ff.render([233, 22139283482834, 3234234])
        ['2.33e2', '2.21e13', '3.23e6']
        >>> ff.render([233, 22139283, 3234234])
        ['233', "22'139'283", "3'234'234"]
        >>> from math import nan, inf
        >>> ff.render([22139283, inf, -inf, nan, None])
        ["22'139'283", 'inf', '-inf', 'nan', None]
        >>> ff.render([1E-4, 1/7, 123456789012345678])
        ['1.00e-4', '1.43e-1', '1.23e17']
        >>> ff.render([0, 0.02, 0.1, 1e-3])
        ['0.000', '0.020', '0.100', '0.001']
        >>> ff.render([-0.2, 1e-6, 0.9])
        ['-2.000e-1', '1.000e-6', '9.000e-1']
        """
        if (source is None) or isinstance(source, (int, float)):
            source = [source]
        if not isinstance(source, Iterable):
            raise type_error(source, "source", Iterable)
        if (none_str is not None) and (
                not isinstance(none_str, FormattedStr)):
            raise type_error(none_str, "none_str", (str, None))

        # get the format parameters
        int_renderer: Final[Callable[[int], str]] = \
            self.get_int_renderer()
        if not callable(int_renderer):
            raise type_error(int_renderer, "int_renderer", call=True)
        int_to_float_threshold: Final[int | float] \
            = self.int_to_float_threshold

        # step one: get the raw numerical data
        data: Final[list[int | float | None]] = \
            cast(list, source) if isinstance(source, list) else list(source)
        dlen: Final[int] = len(data)
        if dlen <= 0:
            raise ValueError("Data cannot be empty.")

        # step two: investigate the data ranges and structure
        all_is_none: bool = True
        all_is_int: bool = True
        max_finite: int | float = -inf
        min_finite: int | float = inf
        min_non_zero_abs: int | float = inf
        longest_fraction: int = -1
        da: int | float

        for i, d in enumerate(data):
            if d is None:
                continue
            all_is_none = False
            d2 = try_int(d) if isfinite(d) else d
            if isinstance(d2, int):
                if d2 < min_finite:
                    min_finite = d2
                if d2 > max_finite:
                    max_finite = d2
                da = abs(d2)
                if 0 < da < min_non_zero_abs:
                    min_non_zero_abs = da
                if not ((-int_to_float_threshold) <= d2
                        <= int_to_float_threshold):
                    d2 = float(d2)
            if d2 is not d:
                data[i] = d2

            if isfinite(d2):
                if not isinstance(d2, int):
                    all_is_int = False
                    s = str(d2)
                    if not (("E" in s) or ("e" in s)):
                        i = s.find(".")
                        if i >= 0:
                            i = len(s) - i - 1
                            if i > longest_fraction:
                                longest_fraction = i
                if d2 < min_finite:
                    min_finite = d2
                if d2 > max_finite:
                    max_finite = d2
                da = abs(d2)
                if 0 < da < min_non_zero_abs:
                    min_non_zero_abs = da

        # step three: if all data is None, we can return here
        if all_is_none:
            return [none_str] * dlen

        # create the protected integer renderer
        def __toint(value: int, form=int_renderer) -> str:
            sv: str = form(value).strip()
            if (sv is not None) and (not isinstance(sv, str)):
                raise type_error(s, f"conversion of {value}", (str, None))
            return sv

        # step four: if all data are integer, we can convert them directly
        if all_is_int:
            # an int render also processing None and special floats
            def __toint2(value: None | int | float, _ns=none_str,
                         form=__toint) -> FormattedStr | None:
                if value is None:
                    return none_str
                return FormattedStr.number(form(cast(int, value))
                                           if isfinite(value) else value)
            return [__toint2(i) for i in data]

        # ok, we have at least some finite floats that cannot be converted to
        # integers. therefore, we need to convert them to strings based on a
        # floating point number format.
        float_format = self.get_float_format(
            min_finite, max_finite, longest_fraction, min_non_zero_abs,
            int_to_float_threshold)
        if not isinstance(float_format, str):
            raise type_error(float_format,
                             "float format from float_format_getter", str)
        if (len(float_format) <= 0) or ("{" not in float_format) \
                or ("}" not in float_format) or (":" not in float_format):
            raise ValueError(f"invalid float format {float_format!r}.")

        def __render_float(value: int | float, ir=__toint,
                           ff=float_format) -> FormattedStr:
            if value is None:
                return none_str
            if isfinite(value):
                res: str = ff.format(value).strip()
                int_part: str
                frac_part: str = ""
                exp_part: str = ""
                eidx: int = res.find("e")
                if eidx < 0:
                    eidx = res.find("E")
                if eidx >= 0:
                    exp_part = f"e{ir(int(res[eidx + 1:])).strip()}"
                    res = res[:eidx].strip()
                dotidx: int = res.find(".")
                if dotidx <= 0:
                    int_part = ir(int(res))
                else:
                    int_part = ir(int(res[:dotidx]))
                    frac_part = res[dotidx:].strip()
                if len(int_part) <= 0:
                    int_part = "0"
                return FormattedStr.number(f"{int_part}{frac_part}{exp_part}")
            return FormattedStr.number(value)

        # step five: first, create the raw float strings and mark special
        # values
        return [__render_float(value) for value in data]


#: the default shared singleton instance of the number group format
DEFAULT_NUMBER_RENDERER: Final[NumberRenderer] = NumberRenderer()



"""
A simple implementation of a (mu+lambda) Evolutionary Algorithm.

This is the basic `mu+lambda`-EA, which works as follows:

1. Start with a list `lst` of `mu` random records and `lambda` blank records.
2. In each iteration:

    2.1. Use the `mu` first records as input to the search operators to
         generate `lambda` new points in the search space.
         For each new point to be created, the binary operator is applied
         with probability `0<=br<=1` and the unary operator is used otherwise.

    2.2. Sort the list `lst` according to the objective value of the record.
         Ties are broken by preferring younger solutions over old ones. Soring
         uses the `__lt__` dunder method of class
         :class:`~moptipy.algorithms.so.record.Record`. This moves the best
         solutions to the front of the list. The tie breaking method both
         encourages drift and ensures compatibility with `RLS`.

If `mu=1`, `lambda=1`, and `br=0`, then this algorithm is exactly equivalent
to the :class:`~moptipy.algorithms.so.rls.RLS` if the same unary and nullary
operator are used. It is only a bit slower due to the additional overhead of
maintaining a list of records. This compatibility is achieved by the tie
breaking strategy of `step 2.2` above: RLS will prefer the newer solution over
the current one if the new solution is either better or as same as good. Now
the latter case cannot be achieved by just sorting the list without
considering the iteration at which a solution was created, since sorting in
Python is *stable* (equal elements remain in the order in which they are
encountered in the original list) and because our new solutions would be in
the `lambda` last entries of the list. This can easily be fixed by the tie
breaking, which is implemented in the `__lt__` dunder method of class
:class:`~moptipy.algorithms.so.record.Record`.

1. Thomas Bäck, David B. Fogel, and Zbigniew Michalewicz, eds., *Handbook of
   Evolutionary Computation.* 1997. Computational Intelligence Library.
   New York, NY, USA: Oxford University Press, Inc. ISBN: 0-7503-0392-1
2. James C. Spall. *Introduction to Stochastic Search and Optimization.*
   Estimation, Simulation, and Control - Wiley-Interscience Series in Discrete
   Mathematics and Optimization, volume 6. 2003. Chichester, West Sussex, UK:
   Wiley Interscience. ISBN: 0-471-33052-3. http://www.jhuapl.edu/ISSO/.
3. Frank Hoffmeister and Thomas Bäck. Genetic Algorithms and Evolution
   Strategies: Similarities and Differences. In Hans-Paul Schwefel and
   Reinhard Männer, *Proceedings of the International Conference on Parallel
   Problem Solving from Nature (PPSN I),* October 1-3, 1990, Dortmund,
   Germany, volume 496 of Lecture Notes in Computer Science, pages 455-469,
   Berlin/Heidelberg, Germany: Springer. ISBN: 978-3-540-54148-6.
   https://doi.org/10.1007/BFb0029787.
"""
from math import isfinite
from typing import Callable, Final, cast

from numpy.random import Generator

from moptipy.algorithms.so.record import Record
from moptipy.api.algorithm import Algorithm2
from moptipy.api.operators import Op0, Op1, Op2
from moptipy.api.process import Process
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.strings import PART_SEPARATOR, num_to_str_for_name
from moptipy.utils.types import type_error


def _int_0(_: int) -> int:
    """
    Return an integer with value `0`.

    :retval 0: always
    """
    return 0


def _float_0() -> float:
    """
    Return a float with value `0.0`.

    :retval 0.0: always
    """
    return 0.0


# start nobinary
class EA(Algorithm2):
    """
    The EA is a population-based algorithm using unary and binary operators.

    It starts with a list of :attr:`mu` randomly initialized solutions. In
    each step, it retains the `mu` best solutions and generates
    :attr:`lambda_` new solutions from them using the unary operator
    (:attr:`~moptipy.api.algorithm.Algorithm1.op1`) with probability
    1-:attr:`br` and the binary search operator
    ((:attr:`~moptipy.api.algorithm.Algorithm2.op2`) at rate :attr:`br`. From
    the joint set of `mu+lambda_` solutions, it again selects the best `mu`
    ones for the next iteration. And so on.

    If `mu=1`, `lambda_=1`, and `br=0`, then this algorithm is exactly
    equivalent to the :class:`~moptipy.algorithms.so.rls.RLS` if the same
    unary and nullary operator are used.
    """

    def solve(self, process: Process) -> None:
        """
        Apply the EA to an optimization problem.

        :param process: the black-box process object
        """
        mu: Final[int] = self.mu  # mu: number of best solutions kept
        mu_plus_lambda: Final[int] = mu + self.lambda_  # size
        # initialization of some variables omitted in book for brevity
        # end nobinary
        random: Final[Generator] = process.get_random()  # random gen
        create: Final[Callable] = process.create  # create x container
        evaluate: Final[Callable] = process.evaluate  # the objective
        op0: Final[Callable] = self.op0.op0  # the nullary operator
        op1: Final[Callable] = self.op1.op1  # the unary operator
        op2: Final[Callable] = self.op2.op2  # the binary operator
        br: Final[float] = self.br  # the rate at which to use op2
        should_terminate: Final[Callable] = process.should_terminate
        r0i: Final[Callable[[int], int]] = cast(   # only if m > 1, we
            Callable[[int], int], random.integers  # need random
            if mu > 1 else _int_0)                 # indices
        r01: Final[Callable[[], float]] = cast(  # only if 0<br<1, we
            Callable[[], float],                 # need random floats
            random.random if 0 < br < 1 else _float_0)
        # start nobinary
        # create list of mu random records and lambda empty records
        lst: Final[list] = [None] * mu_plus_lambda  # pre-allocate list
        f: int | float = 0  # variable to hold objective values
        for i in range(mu_plus_lambda):  # fill list of size mu+lambda
            x = create()  # by creating point in search space
            if i < mu:  # only the first mu records are initialized by
                op0(random, x)  # applying nullary operator = randomize
                if should_terminate():  # should we quit?
                    return   # computational budget exhausted -> quit
                f = evaluate(x)  # continue? ok, evaluate new solution
            lst[i] = Record(x, f)  # create and store record

        it: int = 0  # set iteration counter=0 (but immediately increment)
        while True:  # lst: keep 0..mu-1, overwrite mu..mu+lambda-1
            it += 1  # step iteration counter
            for oi in range(mu, mu_plus_lambda):  # for all offspring
                if should_terminate():  # only continue if we still...
                    return  # have sufficient budget ... otherwise quit
                dest: Record = lst[oi]  # pick destination record
                x = dest.x  # the destination "x" value
                dest.it = it  # remember iteration of solution creation

                sx = lst[r0i(mu)].x  # pick a random source "x"
                # end nobinary
                # start binary
                if r01() < br:  # apply binary operator at rate br
                    sx2 = sx    # second source "x" initially=first sx
                    while sx2 is sx:     # until different from sx...
                        sx2 = lst[r0i(mu)].x  # ..get random second "x"
                    op2(random, x, sx, sx2)  # apply binary operator
                    dest.f = evaluate(x)  # evaluate new point
                    continue  # below is "else" part with unary operat.
                # end binary
                # start nobinary
                op1(random, x, sx)  # apply unary operator
                dest.f = evaluate(x)  # evaluate new point

            lst.sort()  # best records come first, ties broken by age
# end nobinary

    def __init__(self, op0: Op0,
                 op1: Op1 | None = None,
                 op2: Op2 | None = None,
                 mu: int = 1, lambda_: int = 1,
                 br: float | None = None,
                 name: str = "ea") -> None:
        """
        Create the Evolutionary Algorithm (EA).

        :param op0: the nullary search operator
        :param op1: the unary search operator
        :param op2: the binary search operator
        :param mu: the number of best solutions to survive in each generation
        :param lambda_: the number of offspring in each generation
        :param br: the rate at which the binary operator is applied
        :param name: the base name of the algorithm
        """
        if op1 is None:
            op1 = Op1()
            if br is None:
                br = 1.0
            elif br != 1.0:
                raise ValueError(
                    f"if op1==None, br must be None or 1.0, but is {br}.")
        elif op1.__class__ is Op1:
            if br is None:
                br = 1.0
            elif br != 1.0:
                raise ValueError(
                    f"if op1 is Op1, br must be None or 1.0, but is {br}.")
        elif (br is not None) and (br == 1.0):
            op1 = Op1()

        if op2 is None:
            op2 = Op2()
            if br is None:
                br = 0.0
            elif br != 0.0:
                raise ValueError(
                    f"if op2==None, br must be None or 0.0, but is {br}.")
        elif op2.__class__ is Op2:
            if br is None:
                br = 0.0
            elif br != 0.0:
                raise ValueError(
                    f"if op2 is Op2, br must be None or 0.0, but is {br}.")
        elif (br is not None) and (br == 0.0):
            op2 = Op2()
        elif (mu == 1) and (br is None):
            br = 0.0
            op2 = Op2()

        if br is None:
            br = 0.2

        name = f"{name}{PART_SEPARATOR}{mu}{PART_SEPARATOR}{lambda_}"
        if 0 < br < 1:
            name = f"{name}{PART_SEPARATOR}{num_to_str_for_name(br)}"
        super().__init__(name, op0, op1, op2)

        if not isinstance(mu, int):
            raise type_error(mu, "mu", int)
        if not (0 < mu <= 1_000_000):
            raise ValueError(f"invalid mu={mu}, must be in 1..1_000_000.")
        if not isinstance(lambda_, int):
            raise type_error(lambda_, "lambda", int)
        if not (0 < lambda_ <= 1_000_000):
            raise ValueError(
                f"invalid lambda={lambda_}, must be in 1..1_000_000.")
        if not isinstance(br, float):
            raise type_error(br, "br", float)
        if not (isfinite(br) and (0.0 <= br <= 1.0)):
            raise ValueError(f"invalid br={br}, must be in [0, 1].")
        if (br > 0.0) and (mu <= 1):
            raise ValueError(
                f"if br (={br}) > 0, then mu (={mu}) must be > 1.")
        #: the number of records to survive in each generation
        self.mu: Final[int] = mu
        #: the number of offsprings per generation
        self.lambda_: Final[int] = lambda_
        #: the rate at which the binary operator is applied
        self.br: Final[float] = br

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of the algorithm to a logger.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        logger.key_value("mu", self.mu)
        logger.key_value("lambda", self.lambda_)
        logger.key_value("br", self.br)



"""Compute the end result of one mock run."""

from dataclasses import dataclass
from math import ceil, inf
from typing import Final

from numpy.random import Generator

from moptipy.evaluation.end_results import EndResult
from moptipy.mock.components import (
    Algorithm,
    BasePerformance,
    Experiment,
    Instance,
)
from moptipy.utils.console import logger
from moptipy.utils.nputils import rand_generator
from moptipy.utils.types import type_error


def end_result(performance: BasePerformance, seed: int,
               max_fes: int | None = None,
               max_time_millis: int | None = None) -> EndResult:
    """
    Compute the end result of a mock run.

    :param performance: the performance record
    :param seed: the random seed
    :param max_fes: the maximum number of FEs
    :param max_time_millis: the maximum time
    :returns: the end result record
    """
    if not isinstance(performance, BasePerformance):
        raise type_error(performance, "performance", BasePerformance)

    limit_time: int | float = inf
    limit_fes: int | float = inf
    if max_time_millis is not None:
        if not isinstance(max_time_millis, int):
            raise type_error(max_time_millis, "max_time_millis", int)
        if max_time_millis <= 10:
            raise ValueError(
                f"max_time_millis must be > 10, but is {max_time_millis}.")
        limit_time = max_time_millis
    if max_fes is not None:
        if not isinstance(max_fes, int):
            raise type_error(max_fes, "max_fes", int)
        if max_fes <= 10:
            raise ValueError(f"max_fes must be > 10, but is {max_fes}.")
        limit_fes = max_fes

    # The random number generator is determined by the seed.
    random: Final[Generator] = rand_generator(seed)

    # The speed also has some slight jitter.
    jitter: Final[float] = performance.jitter
    speed: float = -1
    while (speed <= 0) or (speed >= 1):
        speed = random.normal(loc=performance.speed, scale=0.01 * jitter)

    # total_time ~ total_fes * (performance.speed ** 3)
    total_time: int
    total_fes: int
    trials: int
    if max_time_millis:
        total_time = int(max_time_millis + abs(random.normal(
            loc=0, scale=5 * jitter)))
        total_fes = -1
        trials = 0
        while ((total_fes <= 100) or (total_fes > limit_fes)) \
                and (trials < 10000):
            trials += 1
            total_fes = int(random.normal(
                loc=max(10.0, total_time / (speed ** 3)),
                scale=max(100.0, 200.0 / speed)))
        if trials >= 10000:
            total_fes = int(min(limit_fes, 10000.0))
    else:
        total_fes = max_fes if max_fes else 1_000_000
        total_time = -1
        trials = 0
        while ((total_time <= 10) or (total_time > limit_time)) \
                and (trials < 10000):
            total_time = int(random.normal(
                loc=max(10.0, total_fes * (speed ** 3)),
                scale=max(10.0, 100.0 / speed)))
        if trials >= 10000:
            total_time = int(min(limit_time, 10000.0))

    # We now look for the vicinity of the local optimum that will be found.
    # We use the quality to determine which attractor to use.
    # Then we will sample a solution between the next lower and next higher
    # attractor, again using the jitter and quality.

    # First, add some jitter to the quality.
    qual: float = -1
    while (qual <= 0) or (qual >= 1):
        qual = random.normal(loc=performance.performance, scale=0.02 * jitter)

    # Second, find the right attractor and remember it in base.
    att: Final[tuple[int, ...]] = performance.instance.attractors
    attn: Final[int] = len(att)
    att_index: int = -1
    best: Final[int] = performance.instance.best
    worst: Final[int] = performance.instance.worst
    while (att_index < 0) or (att_index >= (attn - 1)):
        att_index = int(random.normal(loc=attn * (qual ** 1.7),
                                      scale=jitter ** 0.9))
    base: Final[int] = att[att_index]

    # Third, choose the ends of the intervals in which we can jitter.
    jit_end: int = min(int(base + 0.6 * (att[att_index + 1] - base)), worst)
    jit_start: int = base
    if att_index > 0:
        jit_start = int(0.5 + ceil(base - 0.6 * (base - att[att_index - 1])))
    jit_start = max(jit_start, best)

    # Now determine best_f.
    best_f: int = -1
    while (best_f < jit_start) or (best_f > jit_end) \
            or (best_f < best) or (best_f > worst):
        uni: float = -1
        while (uni <= 0) or (uni >= 1):
            uni = abs(random.normal(loc=0, scale=jitter))
        best_f = int(round(base - uni * (base - jit_start))) \
            if random.uniform(low=0, high=1) < qual else \
            int(round(base + uni * (jit_end - base)))

    # Finally, we need to compute the time we have used.
    fact: float = -1
    while (fact <= 0) or (fact >= 1):
        fact = 1 - random.exponential(scale=(att_index + 1) / (attn + 1))

    last_improvement_fe: int = -1
    while (last_improvement_fe <= 0) or (last_improvement_fe >= total_fes):
        last_improvement_fe = int(random.normal(
            loc=total_fes * fact, scale=total_fes * 0.05 * jitter))

    last_improvement_time: int = -1
    while (last_improvement_time <= 0) \
            or (last_improvement_time >= total_time):
        last_improvement_time = int(random.normal(
            loc=total_time * fact, scale=total_time * 0.05 * jitter))

    res: Final[EndResult] = EndResult(
        algorithm=performance.algorithm.name,
        instance=performance.instance.name,
        rand_seed=seed,
        best_f=best_f,
        last_improvement_fe=last_improvement_fe,
        last_improvement_time_millis=last_improvement_time,
        total_fes=total_fes,
        total_time_millis=total_time,
        goal_f=performance.instance.best,
        max_fes=max_fes,
        max_time_millis=max_time_millis)
    return res


@dataclass(frozen=True, init=False, order=True)
class EndResults:
    """An immutable set of end results."""

    #: The experiment.
    experiment: Experiment
    #: The end results.
    results: tuple[EndResult, ...]
    #: The maximum permitted FEs.
    max_fes: int | None
    #: The maximum permitted milliseconds.
    max_time_millis: int | None
    #: the results per algorithm
    __results_for_algo: dict[str | Algorithm, tuple[EndResult, ...]]
    #: the results per instance
    __results_for_inst: dict[str | Instance, tuple[EndResult, ...]]

    def __init__(self,
                 experiment: Experiment,
                 results: tuple[EndResult, ...],
                 max_fes: int | None = None,
                 max_time_millis: int | None = None):
        """
        Create a mock results of an experiment.

        :param experiment: the experiment
        :param results: the end results
        :param max_fes: the maximum permitted FEs
        :param max_time_millis: the maximum permitted milliseconds.
        """
        if not isinstance(experiment, Experiment):
            raise type_error(experiment, "experiment", Experiment)
        object.__setattr__(self, "experiment", experiment)

        per_algo: Final[dict[str | Algorithm, list[EndResult]]] = {}
        per_inst: Final[dict[str | Instance, list[EndResult]]] = {}
        if not isinstance(results, tuple):
            raise type_error(results, "results", tuple)
        if len(results) <= 0:
            raise ValueError("end_results must not be empty.")
        for a in results:
            if not isinstance(a, EndResult):
                raise type_error(a, "element of results", EndResult)
            aa = experiment.get_algorithm(a.algorithm)
            if aa in per_algo:
                per_algo[aa].append(a)
            else:
                per_algo[aa] = [a]
            ii = experiment.get_instance(a.instance)
            if ii in per_inst:
                per_inst[ii].append(a)
            else:
                per_inst[ii] = [a]

        object.__setattr__(self, "results", results)

        pa: dict[str | Algorithm, tuple[EndResult, ...]] = {}
        for ax in experiment.algorithms:
            lax: list[EndResult] = per_algo[ax]
            lax.sort()
            pa[ax.name] = pa[ax] = tuple(lax)
        pi: dict[str | Instance, tuple[EndResult, ...]] = {}
        for ix in experiment.instances:
            lix: list[EndResult] = per_inst[ix]
            lix.sort()
            pi[ix.name] = pi[ix] = tuple(lix)

        object.__setattr__(self, "_EndResults__results_for_algo", pa)
        object.__setattr__(self, "_EndResults__results_for_inst", pi)

        if max_fes is not None:
            if not isinstance(max_fes, int):
                raise type_error(max_fes, "max_fes", int)
            if max_fes <= 0:
                raise ValueError(f"max_fes must be > 0, but are {max_fes}.")
        object.__setattr__(self, "max_fes", max_fes)

        if max_time_millis is not None:
            if not isinstance(max_time_millis, int):
                raise type_error(max_time_millis, "max_time_millis", int)
            if max_time_millis <= 0:
                raise ValueError("max_time_millis must be > 0, "
                                 f"but are {max_time_millis}.")
        object.__setattr__(self, "max_time_millis", max_time_millis)

    @staticmethod
    def create(experiment: Experiment,
               max_fes: int | None = None,
               max_time_millis: int | None = None) -> "EndResults":
        """
        Create the end results for a given experiment.

        :param experiment: the experiment
        :param max_fes: the maximum number of FEs
        :param max_time_millis: the maximum time
        :returns: the end results
        """
        if not isinstance(experiment, Experiment):
            raise type_error(experiment, "experiment", Experiment)
        logger(
            "now creating all end results for an experiment with "
            f"{len(experiment.algorithms)} algorithms, "
            f"{len(experiment.instances)} instances, and "
            f"{len(experiment.per_instance_seeds[0])} runs per setup.")

        if max_fes is not None:
            if not isinstance(max_fes, int):
                raise type_error(max_fes, "max_fes", int)
            if max_fes <= 0:
                raise ValueError(f"max_fes must be > 0, but are {max_fes}.")

        if max_time_millis is not None:
            if not isinstance(max_time_millis, int):
                raise type_error(max_time_millis, "max_time_millis", int)
            if max_time_millis <= 0:
                raise ValueError("max_time_millis must be > 0, "
                                 f"but are {max_time_millis}.")
        results: list[EndResult] = []
        for per in experiment.applications:
            for seed in experiment.seeds_for_instance(per.instance):
                results.append(end_result(performance=per,
                                          seed=seed,
                                          max_fes=max_fes,
                                          max_time_millis=max_time_millis))
        results.sort()

        res: Final[EndResults] = EndResults(experiment=experiment,
                                            results=tuple(results),
                                            max_fes=max_fes,
                                            max_time_millis=max_time_millis)
        logger(f"finished creating all {len(res.results)} end results.")
        return res

    def results_for_algorithm(self, algorithm: str | Algorithm) \
            -> tuple[EndResult, ...]:
        """
        Get the end results per algorithm.

        :param algorithm: the algorithm
        :returns: the end results
        """
        return self.__results_for_algo[algorithm]

    def results_for_instance(self, instance: str | Instance) \
            -> tuple[EndResult, ...]:
        """
        Get the end results per instance.

        :param instance: the instance
        :returns: the end results
        """
        return self.__results_for_inst[instance]



"""
The base classes for multi-objective optimization algorithms.

A multi-objective optimization algorithm is an optimization algorithm that can
optimize multiple, possibly conflicting, objective functions at once. All such
algorithms inherit from :class:`~moptipy.api.mo_algorithm.MOAlgorithm`. When
developing or implementing new algorithms, you would still follow the concepts
discussed in module :mod:`~moptipy.api.algorithm`.

If you implement a new multi-objective algorithm, you can test it via the
pre-defined unit test routine
:func:`~moptipy.tests.mo_algorithm.validate_mo_algorithm`.
"""
from typing import cast

from moptipy.api.algorithm import Algorithm
from moptipy.api.mo_process import MOProcess
from moptipy.api.process import Process
from moptipy.utils.types import type_error


class MOAlgorithm(Algorithm):
    """A base class for multi-objective optimization algorithms."""

    def solve(self, process: Process) -> None:
        """
        Forward to :meth:`solve_mo` and cast `process` to `MOProcess`.

        :param process: the process to solve. Must be an instance of
            :class:`~moptipy.api.mo_process.MOProcess`.
        """
        if not isinstance(process, MOProcess):
            raise type_error(process, "process", MOProcess)
        self.solve_mo(cast(MOProcess, process))

    def solve_mo(self, process: MOProcess) -> None:
        """
        Apply the multi-objective optimization algorithm to the given process.

        :param process: the multi-objective process which provides methods to
            access the search space, the termination criterion, and a source
            of randomness. It also wraps the objective function, remembers the
            best-so-far solution, and takes care of creating log files (if
            this is wanted).
        """



"""
A set of numerical optimization algorithms from SciPy.

The function :func:`scipy.optimize.minimize` provides a set of very
efficient numerical/continuous optimization methods. Here we wrap a set of
them into our `moptipy` :class:`~moptipy.api.process.Process` API. All
algorithms provided in this module are imported and wrapped from SciPy
(https://scipy.org).

By using the :func:`~moptipy.api.subprocesses.without_should_terminate`
tool, we can enforce the termination criteria set via the
:class:`~moptipy.api.execution.Execution` builder on external algorithms
while piping all their function evaluations through the
:meth:`~moptipy.api.process.Process.evaluate` routine of the optimization
:meth:`~moptipy.api.process.Process`. This way, we can make these external
algorithms usable within `moptipy` in a transparent manner.
"""
from typing import Any, Callable, Final, cast

import numpy as np
from numpy import ndarray
from scipy.optimize import Bounds  # type: ignore

# isort: off
# noinspection PyProtectedMember
from scipy.optimize._differentialevolution import (  # type: ignore
    differential_evolution,  # type: ignore
)
# isort: on

# noinspection PyProtectedMember
from scipy.optimize._optimize import (  # type: ignore
    _minimize_bfgs,  # type: ignore
    _minimize_cg,  # type: ignore
    _minimize_neldermead,  # type: ignore
    _minimize_powell,  # type: ignore
)

# noinspection PyProtectedMember
from scipy.optimize._slsqp_py import _minimize_slsqp  # type: ignore

# noinspection PyProtectedMember
from scipy.optimize._tnc import _minimize_tnc  # type: ignore

from moptipy.api.algorithm import Algorithm, Algorithm0
from moptipy.api.operators import Op0
from moptipy.api.process import Process
from moptipy.api.subprocesses import (
    get_remaining_fes,
    without_should_terminate,
)
from moptipy.spaces.vectorspace import VectorSpace
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.types import type_error


class SciPyAlgorithmWrapper(Algorithm0):
    """
    A wrapper for the Sci-Py API.

    An instance of this class may be re-used, but it must only be used for
    problems of the same dimension.
    """

    def __init__(self, name: str, op0: Op0, space: VectorSpace) -> None:
        """
        Create the algorithm importer from scipy.

        :param name: the name of the algorithm
        :param op0: the nullary search operator
        :param space: the vector space
        """
        super().__init__(name, op0)
        if not isinstance(space, VectorSpace):
            raise type_error(space, "space", VectorSpace)
        #: the vector space defining the dimensions and bounds
        self.space: Final[VectorSpace] = space
        #: the bounds to be used for the internal function call
        self.__bounds: Final[Bounds] = Bounds(space.lower_bound,
                                              space.upper_bound)
        #: the cache for starting points
        self.__x0: Final[ndarray] = space.create()

    def _call(self, func: Callable[[np.ndarray], int | float],
              x0: np.ndarray, max_fes: int, bounds: Bounds) -> None:
        """
        Invoke the SciPi Algorithm.

        This function will be overwritten to call the SciPi Algorithm.

        :param func: the function to minimize
        :param x0: the starting point
        :param max_fes: the maximum FEs
        :param bounds: the bounds
        """

    def __run(self, process: Process) -> None:
        """
        Execute the algorithm.

        :param process: the process
        """
        x0: Final[np.ndarray] = self.__x0
        self.op0.op0(process.get_random(), x0)  # create first solution
        self._call(self.space.clipped(process.evaluate),
                   x0, get_remaining_fes(process),
                   self.__bounds)  # invoke the algorithm

    def solve(self, process: Process) -> None:
        """
        Apply the algorithm from SciPy to an optimization problem.

        Basically, this wraps a specific configuration of
        :func:`scipy.optimize.minimize` into our process API and
        invokes it.

        :param process: the black-box process object
        """
        # invoke the SciPy algorithm implementation
        without_should_terminate(
            cast(Callable[[Process], Any], self.__run), process)

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of the algorithm to a logger.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)  # log algorithm/operator
        self.space.log_bounds(logger)  # log bounds


# noinspection PyProtectedMember
def _call_powell(func: Callable[[np.ndarray], int | float],
                 x0: np.ndarray, max_fes: int, bounds: Bounds) -> None:
    _minimize_powell(func, x0, bounds=bounds, xtol=0.0, ftol=0.0,
                     maxiter=max_fes, maxfev=max_fes)


class Powell(SciPyAlgorithmWrapper):
    """
    Powell's Algorithm.

    The function :func:`scipy.optimize.minimize` with parameter
    "Powell" for continuous optimization.

    1. Michael James David Powell. An Efficient Method for Finding the Minimum
       of a Function of Several Variables without Calculating Derivatives. The
       Computer Journal. 7(2):155-162. 1964.
       https://doi.org/10.1093/comjnl/7.2.155
    """

    def __init__(self, op0: Op0, space: VectorSpace) -> None:
        """
        Create Powell's algorithm importer from scipy.

        :param op0: the nullary search operator
        :param space: the vector space
        """
        super().__init__("powell_scipy", op0, space)
        self._call = _call_powell  # type: ignore


def _call_nelder_mead(func: Callable[[np.ndarray], int | float],
                      x0: np.ndarray, max_fes: int, bounds: Bounds) -> None:
    _minimize_neldermead(func, x0, bounds=bounds, xatol=0.0, fatol=0.0,
                         maxiter=max_fes, maxfev=max_fes)


class NelderMead(SciPyAlgorithmWrapper):
    """
    The Downhill Simplex aka. the Nelder-Mead Algorithm.

    The function :func:`scipy.optimize.minimize` with parameter
    "Nelder-Mead" for continuous optimization  by using the Downhill Simplex
    algorithm a.k.a., the Nelder-Mead algorithm. Here we wrap it into our API.

    Scipy provides the following reference:

    1. Fuchang Gao and Lixing Han. Implementing the Nelder-Mead Simplex
       Algorithm with Adaptive Parameters. *Computational Optimization and
       Applications*. 51(1):259-277. January 2012.
       https://doi.org/10.1007/s10589-010-932
    2. J. A. Nelder and R. Mead. A Simplex Method for Function Minimization.
       *The Computer Journal*. 7(4):308-313. January 1965. Oxford University
       Press (OUP). http://dx.doi.org/10.1093/COMJNL/7.4.308
       https://people.duke.edu/~hpgavin/cee201/Nelder+Mead-\
ComputerJournal-1965.pdf
    3. M. H. Wright. Direct Search Methods: Once Scorned, Now Respectable.
       In D.F. Griffiths and G.A. Watson (Eds.) *Proceedings of the 1995
       Dundee Biennial Conference in Numerical Analysis*. Harlow, UK:
       Addison Wesley Longman, pp. 191-208.
    4. Jorge Nocedal and Stephen J. Wright. *Numerical Optimization*. In
       Springer Series in Operations Research and Financial Engineering.
       New York, NY, USA: Springer. 2006. Second Edition.
       ISBN: 978-0-387-30303-1. Chapter 9.5, Page 238.
       https://doi.org/10.1007/978-0-387-40065-5.
    """

    def __init__(self, op0: Op0, space: VectorSpace) -> None:
        """
        Create the Nelder-Mead Downhill Simplex importer from scipy.

        :param op0: the nullary search operator
        :param space: the vector space
        """
        super().__init__("nelderMead_scipy", op0, space)
        self._call = _call_nelder_mead  # type: ignore


def _call_bgfs(func: Callable[[np.ndarray], int | float],
               x0: np.ndarray, max_fes: int, _) -> None:
    _minimize_bfgs(func, x0, gtol=0.0, maxiter=max_fes)


class BGFS(SciPyAlgorithmWrapper):
    """
    The wrapper for the BGFS algorithm in SciPy.

    This is the quasi-Newton method by C. G. Broyden, Roger Fletcher,
    D. Goldfarb, and David F. Shanno (BFGS).

    1. Jorge Nocedal and Stephen J. Wright. *Numerical Optimization*. In
       Springer Series in Operations Research and Financial Engineering.
       New York, NY, USA: Springer. 2006. Second Edition.
       ISBN: 978-0-387-30303-1. Chapter 6, Page 136.
       https://doi.org/10.1007/978-0-387-40065-5.
    2. Roger Fletcher. *Practical Methods of Optimization* (2nd ed.),
       New York: John Wiley & Sons. 1987. ISBN 978-0-471-91547-8.
    3. C. G. Broyden. The convergence of a class of double-rank minimization
       algorithms. *Journal of the Institute of Mathematics and Its
       Applications*. 6(1):76-90. March 1970.
       http://dx.doi.org/10.1093/imamat/6.1.76
    """

    def __init__(self, op0: Op0, space: VectorSpace) -> None:
        """
        Create BGFS algorithm importer from scipy.

        :param op0: the nullary search operator
        :param space: the vector space
        """
        super().__init__("bgfs_scipy", op0, space)
        self._call = _call_bgfs  # type: ignore


def _call_cg(func: Callable[[np.ndarray], int | float],
             x0: np.ndarray, max_fes: int, _) -> None:
    _minimize_cg(func, x0, gtol=0.0, maxiter=max_fes)


class CG(SciPyAlgorithmWrapper):
    """
    The wrapper for the Conjugate Gradient algorithm in SciPy.

    1. Jorge Nocedal and Stephen J. Wright. *Numerical Optimization*. In
       Springer Series in Operations Research and Financial Engineering.
       New York, NY, USA: Springer. 2006. Second Edition.
       ISBN: 978-0-387-30303-1. Chapter 5, Page 101.
    """

    def __init__(self, op0: Op0, space: VectorSpace) -> None:
        """
        Create Conjugate Gradient algorithm importer from scipy.

        :param op0: the nullary search operator
        :param space: the vector space
        """
        super().__init__("cg_scipy", op0, space)
        self._call = _call_cg  # type: ignore


def _call_slsqp(func: Callable[[np.ndarray], int | float],
                x0: np.ndarray, max_fes: int, _) -> None:
    _minimize_slsqp(func, x0, ftol=0.0, maxiter=max_fes)


class SLSQP(SciPyAlgorithmWrapper):
    """
    The Sequential Least Squares Programming (SLSQP) algorithm in SciPy.

    1. Dieter Kraft. Algorithm 733: TOMP-Fortran modules for optimal control
       calculations. *ACM Transactions on Mathematical Software.*
       20(3):262-281. September 1994. https://doi.org/10.1145/192115.192124
    """

    def __init__(self, op0: Op0, space: VectorSpace) -> None:
        """
        Create the SLSQP algorithm importer from scipy.

        :param op0: the nullary search operator
        :param space: the vector space
        """
        super().__init__("slsqp_scipy", op0, space)
        self._call = _call_slsqp  # type: ignore


def _call_tnc(func: Callable[[np.ndarray], int | float],
              x0: np.ndarray, max_fes: int, bounds: Bounds) -> None:
    _minimize_tnc(
        func, x0,
        bounds=[(lb, bounds.ub[i]) for i, lb in enumerate(bounds.lb)],
        ftol=0.0, xtol=0.0, gtol=0.0, maxiter=max_fes, maxfun=max_fes)


class TNC(SciPyAlgorithmWrapper):
    """
    The Truncated Newton Method from SciPy.

    1. Stephen G. Nash. Newton-Type Minimization via the Lanczos Method.
       *SIAM Journal on Numerical Analysis*. 21(4):770-783. August 1984.
       https://dx.doi.org/10.1137/0721052.
    2. Jorge Nocedal and Stephen J. Wright. *Numerical Optimization*. In
       Springer Series in Operations Research and Financial Engineering.
       New York, NY, USA: Springer. 2006. Second Edition.
       ISBN: 978-0-387-30303-1. https://doi.org/10.1007/978-0-387-40065-5.
    """

    def __init__(self, op0: Op0, space: VectorSpace) -> None:
        """
        Create the TNC algorithm importer from scipy.

        :param op0: the nullary search operator
        :param space: the vector space
        """
        super().__init__("tnc_scipy", op0, space)
        self._call = _call_tnc  # type: ignore


class DE(Algorithm):
    """
    The Differential Evolution Algorithm as implemented by SciPy.

    At this point, we do not expose the many parameters of the function
    :func:`scipy.optimize.differential_evolution`.
    We only use the default settings. This may change in future releases.

    1. Rainer Storn and Kenneth Price. Differential Evolution - A Simple and
       Efficient Heuristic for global Optimization over Continuous Spaces.
       *Journal of Global Optimization* 11(4):341-359. December 1997.
       https://doi.org/10.1023/A:1008202821328.
       https://www.researchgate.net/publication/227242104
    """

    def __init__(self, space: VectorSpace) -> None:
        """
        Create the Differential Evolution Algorithm from SciPy.

        :param space: the vector space
        """
        super().__init__()
        if not isinstance(space, VectorSpace):
            raise type_error(space, "space", VectorSpace)
        #: the vector space defining the dimensions and bounds
        self.space: Final[VectorSpace] = space
        #: the bounds of the search space, derived from :attr:`space`
        self.__bounds: Final[list[tuple[float, float]]] = \
            [(lb, space.upper_bound[i])
             for i, lb in enumerate(space.lower_bound)]

    def __run(self, process: Process) -> None:
        """
        Execute the algorithm.

        :param process: the process
        """
        mf = get_remaining_fes(process)  # get the number of available FEs

        differential_evolution(
            self.space.clipped(process.evaluate),
            bounds=self.__bounds,
            maxiter=int(mf / len(self.__bounds)) + 1,
            tol=0.0, seed=process.get_random(), atol=0.0)

    def solve(self, process: Process) -> None:
        """
        Apply the algorithm from SciPy to an optimization problem.

        Basically, this wraps a specific configuration of
        :func:`scipy.optimize.minimize` into our process API and
        invokes it.

        :param process: the black-box process object
        """
        # invoke the SciPy algorithm implementation
        without_should_terminate(
            cast(Callable[[Process], Any], self.__run), process)

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of the algorithm to a logger.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)  # log algorithm/operator
        with logger.scope("space") as sp:
            self.space.log_parameters_to(sp)  # log space

    def __str__(self):
        """
        Get the name of this algorithm.

        :returns: the name of this differential evolution algorithm
        """
        return "de_scipy"



"""Functions that can be used to test multi-objective algorithms."""
from math import inf, isfinite
from typing import Any, Final

from numpy import array_equal
from numpy.random import Generator, default_rng

from moptipy.api.algorithm import (
    Algorithm0,
    Algorithm1,
    Algorithm2,
    check_algorithm,
)
from moptipy.api.encoding import Encoding
from moptipy.api.mo_algorithm import MOAlgorithm
from moptipy.api.mo_archive import MOArchivePruner
from moptipy.api.mo_execution import MOExecution
from moptipy.api.mo_problem import MOProblem
from moptipy.api.operators import check_op0, check_op1, check_op2
from moptipy.api.space import Space
from moptipy.mo.archive.keep_farthest import KeepFarthest
from moptipy.tests.component import validate_component
from moptipy.tests.encoding import validate_encoding
from moptipy.tests.mo_problem import validate_mo_problem
from moptipy.tests.space import validate_space
from moptipy.utils.types import type_error


def validate_mo_algorithm(
        algorithm: MOAlgorithm,
        solution_space: Space,
        problem: MOProblem,
        search_space: Space | None = None,
        encoding: Encoding | None = None,
        max_fes: int = 100,
        is_encoding_deterministic: bool = True) -> None:
    """
    Check whether a multi-objective algorithm follows the moptipy API.

    :param algorithm: the algorithm to test
    :param solution_space: the solution space
    :param problem: the problem to solve
    :param search_space: the optional search space
    :param encoding: the optional encoding
    :param max_fes: the maximum number of FEs
    :param is_encoding_deterministic: is the encoding deterministic?
    :raises TypeError: if `algorithm` is not a
        :class:`~moptipy.api.mo_algorithm.MOAlgorithm` instance
    :raises ValueError: if `algorithm` does not behave like it should
    """
    if not isinstance(algorithm, MOAlgorithm):
        raise type_error(algorithm, "algorithm", MOAlgorithm)

    check_algorithm(algorithm)
    if isinstance(algorithm, Algorithm0):
        check_op0(algorithm.op0)
        if isinstance(algorithm, Algorithm1):
            check_op1(algorithm.op1)
            if isinstance(algorithm, Algorithm2):
                check_op2(algorithm.op2)

    validate_component(algorithm)
    validate_mo_problem(problem, None, None)
    validate_space(solution_space, None)

    if encoding is not None:
        validate_encoding(encoding, None, None, None,
                          is_encoding_deterministic)
        validate_space(search_space, None)

    if not isinstance(max_fes, int):
        raise type_error(max_fes, "max_fes", int)
    if max_fes <= 0:
        raise ValueError(f"max_fes must be > 0, but is {max_fes}.")

    lb: Final[int | float] = problem.lower_bound()
    if (not isfinite(lb)) and (lb != -inf):
        raise ValueError(f"objective lower bound cannot be {lb}.")
    ub = problem.upper_bound()
    if (not isfinite(ub)) and (ub != inf):
        raise ValueError(f"objective upper bound cannot be {ub}.")

    exp = MOExecution()
    exp.set_algorithm(algorithm)
    exp.set_max_fes(max_fes)
    exp.set_solution_space(solution_space)
    exp.set_objective(problem)
    if search_space is not None:
        exp.set_search_space(search_space)
        exp.set_encoding(encoding)

    random: Final[Generator] = default_rng()
    max_archive_size: Final[int] = int(random.integers(
        1, 1 << int(random.integers(1, 6))))
    exp.set_archive_max_size(max_archive_size)
    exp.set_archive_pruning_limit(
        max_archive_size + int(random.integers(0, 8)))
    if random.integers(2) <= 0:
        choice: int = int(random.integers(2))
        pruner: MOArchivePruner
        if choice <= 0:
            lst: list[int]
            while True:
                lst = [i for i in range(problem.f_dimension())
                       if random.integers(2) <= 0]
                if len(lst) > 0:
                    break
            pruner = KeepFarthest(problem, lst)
        else:
            pruner = MOArchivePruner()
        exp.set_archive_pruner(pruner)

    with exp.execute() as process:
        # re-raise any exception that was caught
        if hasattr(process, "_caught"):
            error = getattr(process, "_caught")
            if error is not None:
                raise error
        # no exception? ok, let's check the data
        if not process.has_best():
            raise ValueError("The algorithm did not produce any solution.")

        if not process.should_terminate():
            raise ValueError("The algorithm stopped before hitting the "
                             "termination criterion.")

        consumed_fes: Final[int] = process.get_consumed_fes()
        if not isinstance(consumed_fes, int):
            raise type_error(consumed_fes, "consumed_fes", int)
        if (consumed_fes <= 0) or (consumed_fes > max_fes):
            raise ValueError(
                f"Consumed FEs must be positive and <= {max_fes}, "
                f"but is {consumed_fes}.")

        last_imp_fe: Final[int] = process.get_last_improvement_fe()
        if not isinstance(last_imp_fe, int):
            raise type_error(last_imp_fe, "last improvement FE", int)
        if (last_imp_fe <= 0) or (last_imp_fe > consumed_fes):
            raise ValueError("Last improvement FEs must be positive and "
                             f"<= {consumed_fes}, but is {last_imp_fe}.")

        consumed_time: Final[int] = process.get_consumed_time_millis()
        if not isinstance(consumed_time, int):
            raise type_error(consumed_time, "consumed time", int)
        if consumed_time < 0:
            raise ValueError(
                f"Consumed time must be >= 0, but is {consumed_time}.")

        last_imp_time: Final[int] = process.get_last_improvement_time_millis()
        if not isinstance(last_imp_time, int):
            raise type_error(last_imp_time, "last improvement time", int)
        if (last_imp_time < 0) or (last_imp_time > consumed_time):
            raise ValueError(
                f"Consumed time must be >= 0 and <= {consumed_time}, but "
                f"is {last_imp_time}.")

        if lb != process.lower_bound():
            raise ValueError(
                "Inconsistent lower bounds between process "
                f"({process.lower_bound()}) and scalarized objective ({lb}).")
        if ub != process.upper_bound():
            raise ValueError(
                "Inconsistent upper bounds between process "
                f"({process.upper_bound()}) and scalarized objective ({ub}).")

        res_f: Final[float | int] = process.get_best_f()
        if not isfinite(res_f):
            raise ValueError("Infinite scalarized objective value of result.")
        if (res_f < lb) or (res_f > ub):
            raise ValueError(
                f"Objective value {res_f} outside of bounds [{lb},{ub}].")

        y = solution_space.create()
        process.get_copy_of_best_y(y)
        solution_space.validate(y)
        fs1 = problem.f_create()
        fs2 = problem.f_create()
        process.get_copy_of_best_y(y)
        check_f = problem.f_evaluate(y, fs1)
        if check_f != res_f:
            raise ValueError(
                f"Inconsistent objective value {res_f} from process compared "
                f"to {check_f} from objective function.")
        process.get_copy_of_best_fs(fs2)
        if not array_equal(fs1, fs2):
            raise ValueError(
                f"Inconsistent objective vectors {fs1} and {fs2}.")

        x: Any | None = None
        if search_space is not None:
            x = search_space.create()
            process.get_copy_of_best_x(x)
            search_space.validate(x)

        if encoding is not None:
            y2 = solution_space.create()
            encoding.decode(x, y2)
            solution_space.validate(y2)
            if is_encoding_deterministic:
                solution_space.is_equal(y, y2)



"""
A nullary operator forwarding to another function.

This is a nullary operator (an instance of
:class:`~moptipy.api.operators.Op0`) whose method
:meth:`~moptipy.api.operators.Op0.op0` forwards to another `Callable`.
This other `Callable` can then return a solution that is created in some
special way, or maybe even the current best solution of a search process.

This operator has been designed to be used in conjunction with
:func:`~moptipy.api.subprocesses.from_starting_point`, which is an
optimization :class:`~moptipy.api.process.Process` where a starting point
has been defined, i.e., where the methods
:meth:`~moptipy.api.process.Process.get_copy_of_best_x` and
:meth:`~moptipy.api.process.Process.get_best_f` return pre-defined values.
By setting :meth:`~moptipy.operators.op0_forward.Op0Forward.forward_to` to
:meth:`~moptipy.api.process.Process.get_copy_of_best_x`, this nullary operator
will return the current-best solution of the optimization process, which, in
this case, will be the pre-defined starting point.
Any optimization algorithm (e.g., an instance of
:class:`~moptipy.api.algorithm.Algorithm0`) using this nullary operator to get
its initial solution will then begin the search at this pre-defined starting
point. This allows using one algorithm as a sub-algorithm of another one.
Wrapping :func:`~moptipy.api.subprocesses.from_starting_point` around the
result of a call to :func:`~moptipy.api.subprocesses.for_fes` would allow to
limit the number of objective function evaluations consumed by the
sub-algorithm.
"""
from typing import Any, Callable

import numpy as np
from numpy.random import Generator

from moptipy.api.operators import Op0
from moptipy.utils.types import type_error


class Op0Forward(Op0):
    """A nullary operator that forwards all calls to `op0` to a `Callable`."""

    def __init__(self):
        """Initialize this operator."""
        #: the internal blueprint for filling permutations
        self.__call: Callable[[Any], None] | None = None

    def op0(self, random: Generator, dest: np.ndarray) -> None:
        """
        Forward the call.

        :param random: ignored
        :param dest: the destination data structure to be filled with the data
            of the point in the search space by the internal `Callable` set by
            :meth:`forward_to`.
        """
        self.__call(dest)

    def forward_to(self, call: Callable[[Any], None]) -> None:
        """
        Set the `Callable` to forward all calls from :meth:`op0` to.

        :param call: the `Callable` to which all calls to :meth:`op0` should
            be delegated to.
        """
        if not callable(call):
            raise type_error(call, "call", call=True)
        self.__call = call

    def stop_forwarding(self) -> None:
        """Stop forwarding the call."""
        self.__call = None

    def __str__(self) -> str:
        """
        Get the name of this operator.

        :return: "forward"
        """
        return "forward"

    def initialize(self) -> None:
        """Initialize this operator by stopping to forward."""
        super().initialize()
        self.stop_forwarding()



"""A set of utilities interactions with the file system."""
import os
from contextlib import AbstractContextManager
from shutil import rmtree
from tempfile import mkdtemp, mkstemp

from moptipy.utils.path import Path


class TempDir(Path, AbstractContextManager):
    """
    A scoped temporary directory to be used in a 'with' block.

    The directory and everything in it will be deleted upon exiting the
    'with' block.
    """

    #: is the directory open?
    __is_open: bool

    def __new__(cls, value):  # noqa
        """
        Construct the object.

        :param value: the string value
        """
        ret = super().__new__(cls, value)
        ret.enforce_dir()
        ret.__is_open = True
        return ret

    @staticmethod
    def create(directory: str | None = None) -> "TempDir":
        """
        Create the temporary directory.

        :param directory: an optional root directory
        :raises TypeError: if `directory` is not `None` but also no `str`
        """
        if directory is not None:
            root_dir = Path.path(directory)
            root_dir.enforce_dir()
        else:
            root_dir = None
        return TempDir(mkdtemp(dir=root_dir))

    def __enter__(self) -> "TempDir":
        """Nothing, just exists for `with`."""
        if not self.__is_open:
            raise ValueError(f"Temporary directory {self!r} already closed.")
        return self

    def __exit__(self, exception_type, exception_value, traceback) -> bool:
        """
        Delete the temporary directory and everything in it.

        :param exception_type: ignored
        :param exception_value: ignored
        :param traceback: ignored
        :returns: `True` to suppress an exception, `False` to rethrow it
        """
        opn = self.__is_open
        self.__is_open = False
        if opn:
            rmtree(self, ignore_errors=True, onerror=None)
        return exception_type is None


class TempFile(Path, AbstractContextManager):
    """
    A scoped temporary file to be used in a 'with' block.

    This file will be deleted upon exiting the 'with' block.
    """

    #: is the directory open?
    __is_open: bool

    def __new__(cls, value):  # noqa
        """
        Construct the object.

        :param value: the string value
        """
        ret = super().__new__(cls, value)
        ret.enforce_file()
        ret.__is_open = True
        return ret

    @staticmethod
    def create(directory: str | None = None,
               prefix: str | None = None,
               suffix: str | None = None) -> "TempFile":
        """
        Create a temporary file.

        :param directory: a root directory or `TempDir` instance
        :param prefix: an optional prefix
        :param suffix: an optional suffix, e.g., `.txt`
        :raises TypeError: if any of the parameters does not fulfill the type
            contract
        """
        if prefix is not None:
            prefix = prefix.strip()
            if not prefix:
                raise ValueError("Prefix cannot be empty if specified.")

        if suffix is not None:
            suffix = suffix.strip()
            if not suffix:
                raise ValueError("Prefix cannot be empty if specified.")

        if directory is not None:
            base_dir = Path.path(directory)
            base_dir.enforce_dir()
        else:
            base_dir = None

        (handle, path) = mkstemp(suffix=suffix, prefix=prefix, dir=base_dir)
        os.close(handle)
        return TempFile(path)

    def __enter__(self) -> "TempFile":
        """Nothing, just exists for `with`."""
        if not self.__is_open:
            raise ValueError(f"Temporary file {self!r} already deleted.")
        return self

    def __exit__(self, exception_type, exception_value, traceback) -> bool:
        """
        Delete the temporary file.

        :param exception_type: ignored
        :param exception_value: ignored
        :param traceback: ignored
        :returns: `True` to suppress an exception, `False` to rethrow it
        """
        opn = self.__is_open
        self.__is_open = False
        if opn:
            os.remove(self)
        return exception_type is None



"""Some basic type handling routines."""
from typing import Any, Final, Iterable


def type_name(tpe: type) -> str:
    """
    Convert a type to a string.

    :param tpe: the type
    :returns: the string

    >>> type_name(type(None))
    'None'
    >>> type_name(int)
    'int'
    >>> from moptipy.utils.path import Path
    >>> type_name(Path)
    'moptipy.utils.path.Path'
    """
    c1: str = str(tpe)
    if c1.startswith("<class '"):
        c1 = c1[8:-2]
    if c1 == "NoneType":
        return "None"

    if hasattr(tpe, "__qualname__"):
        c2: str = tpe.__qualname__
        if hasattr(tpe, "__module__"):
            module = tpe.__module__
            if (module is not None) and (module != "builtins"):
                c2 = f"{module}.{c2}"
        if len(c2) > len(c1):
            return c2
    return c1


def type_name_of(obj) -> str:
    """
    Get the fully-qualified class name of an object.

    :param obj: the object
    :returns: the fully-qualified class name of the object

    >>> from moptipy.examples.jssp.instance import Instance
    >>> type_name_of(Instance.from_resource("demo"))
    'moptipy.examples.jssp.instance.Instance'
    >>> from numpy.random import default_rng
    >>> type_name_of(default_rng())
    'numpy.random._generator.Generator'
    """
    if obj is None:
        return "None"
    c1: Final[str] = type_name(type(obj))
    if hasattr(obj, "__class__"):
        cls: Final[type] = obj.__class__
        c2: str = type_name(cls)

        if hasattr(cls, "__qualname__"):
            c3: str = cls.__qualname__
            if hasattr(obj, "__module__"):
                module = obj.__module__
                if (module is not None) and (module != "builtins"):
                    c3 = f"{module}.{c3}"
            if len(c3) > len(c2):
                c2 = c3

        if len(c2) > len(c1):
            return c2
    return c1


def type_error(obj: Any, name: str,
               expected: None | type | Iterable[type] = None,
               call: bool = False) -> ValueError | TypeError:
    """
    Create an error to raise if a type did not fit.

    :param obj: the object that is of the wrong type
    :param name: the name of the object
    :param expected: the expected types (or `None`)
    :param call: the object should have been callable?
    :returns: a :class:`TypeError` with a descriptive information

    >>> type_error(1.3, "var", int)
    TypeError("var should be an instance of int but is float, namely '1.3'.")
    >>> type_error("x", "z", (int, float)).args[0]
    "z should be an instance of any in {float, int} but is str, namely 'x'."
    >>> type_error("f", "q", call=True).args[0]
    "q should be a callable but is str, namely 'f'."
    >>> type_error("1", "2", bool, call=True).args[0]
    "2 should be an instance of bool or a callable but is str, namely '1'."
    >>> type_error(None, "x", str)
    TypeError('x should be an instance of str but is None.')
    """
    exp: str = ""
    if isinstance(expected, Iterable):
        exp = ", ".join(sorted([type_name(e) for e in expected]))
        exp = f"an instance of any in {{{exp}}}"
    elif expected is not None:
        exp = f"an instance of {type_name(expected)}"
    if call:
        exp = f"{exp} or a callable" if exp else "a callable"

    if obj is None:
        return TypeError(f"{name} should be {exp} but is None.")
    return TypeError(f"{name} should be {exp} but is "
                     f"{type_name_of(obj)}, namely {str(obj)!r}.")



"""
A fitness representing the rank of a solution based on its quality.

First, all solutions are sorted based on their objective values. Then, they
receive their rank, i.e., index in the sorted list, as fitness.

Together with fitness proportionate selection (:class:`~moptipy.algorithms.\
modules.selections.fitness_proportionate_sus.FitnessProportionateSUS`), it
works very similar to linear ranking selection in an EA
(:class:`~moptipy.algorithms.so.general_ea.GeneralEA`).

1. L. Darrell Whitley. The GENITOR Algorithm and Selection Pressure: Why
   Rank-Based Allocation of Reproductive Trials is Best. In J. David Schaffer,
   ed., Proceedings of the 3rd International Conference on Genetic Algorithms
   (ICGA'89), June 4-7, 1989, Fairfax, VA, USA, pages 116-121. San Francisco,
   CA, USA: Morgan Kaufmann Publishers Inc. ISBN: 1-55860-066-3
   https://www.researchgate.net/publication/2527551
2. Tobias Blickle and Lothar Thiele. A Comparison of Selection Schemes used in
   Genetic Algorithms. Second edition, December 1995. TIK-Report 11 from the
   Eidgenössische Technische Hochschule (ETH) Zürich, Department of Electrical
   Engineering, Computer Engineering and Networks Laboratory (TIK), Zürich,
   Switzerland. ftp://ftp.tik.ee.ethz.ch/pub/publications/TIK-Report11.ps
"""

from math import inf

from numpy.random import Generator

from moptipy.algorithms.so.fitness import Fitness, FRecord


# start book
class Rank(Fitness):
    """A fitness computing the rank of an individual based on its quality."""

    def assign_fitness(self, p: list[FRecord], random: Generator) -> None:
        """
        Assign the rank fitness.

        :param p: the list of records
        :param random: ignored

        >>> l = [FRecord(0, 10), FRecord(1, 5), FRecord(2, 5), FRecord(3, -1)]
        >>> from numpy.random import default_rng
        >>> Rank().assign_fitness(l, default_rng())
        >>> l[0].x
        3
        >>> l[0].fitness
        1
        >>> l[1].x
        1
        >>> l[1].fitness
        2
        >>> l[2].x
        2
        >>> l[2].fitness
        2
        >>> l[3].x
        0
        >>> l[3].fitness
        4
        """
        for rec in p:  # first copy rec.f to rec.fitness
            rec.fitness = rec.f  # because then we can easily sort
        p.sort()  # sort based on objective values

        rank: int = -1  # the variable for storing the current rank
        last_f: int | float = -inf  # the previous objective value
        for i, rec in enumerate(p):  # iterate over list
            v = rec.fitness  # get the current objective value
            if v > last_f:  # only increase rank if objective f changes
                rank = i + 1  # +1 so smallest-possible fitness is 1
                last_f = v  # remember objective value for comparison
            rec.fitness = rank  # assign the rank (same f = same rank)
# end book

    def __str__(self):
        """
        Get the name of this default fitness assignment.

        :return: the name of this fitness assignment strategy
        :retval "rank": always
        """
        return "rank"



"""
The Frequency Fitness Assignment (FFA) Process.

Frequency Fitness Assignment (FFA) replaces all objective values with their
encounter frequencies in the selection decisions. The more often an
objective value is encountered, the higher gets its encounter frequency.
Therefore, local optima are slowly receiving worse and worse fitness.

Notice that this implementation of FFA has a slight twist to it:
It will incorporate the iteration index
(:attr:`~moptipy.algorithms.so.record.Record.it`) of the solutions
into the fitness.
This index is used to break ties, in which case newer solutions are preferred.
This can make the EA with FFA compatible with the
:class:`moptipy.algorithms.so.fea1plus1.FEA1plus1` if "best" selection
(:class:`moptipy.algorithms.modules.selections.best.Best`) is used
at mu=lambda=1.

1. Thomas Weise, Zhize Wu, Xinlu Li, and Yan Chen. Frequency Fitness
   Assignment: Making Optimization Algorithms Invariant under Bijective
   Transformations of the Objective Function Value. *IEEE Transactions on
   Evolutionary Computation* 25(2):307-319. April 2021. Preprint available at
   arXiv:2001.01416v5 [cs.NE] 15 Oct 2020.
   https://dx.doi.org/10.1109/TEVC.2020.3032090
2. Thomas Weise, Zhize Wu, Xinlu Li, Yan Chen, and Jörg Lässig. Frequency
   Fitness Assignment: Optimization without Bias for Good Solutions can be
   Efficient. *IEEE Transactions on Evolutionary Computation (TEVC)*. 2022.
   Early Access. https://dx.doi.org/10.1109/TEVC.2022.3191698
3. Thomas Weise, Mingxu Wan, Ke Tang, Pu Wang, Alexandre Devert, and Xin
   Yao. Frequency Fitness Assignment. *IEEE Transactions on Evolutionary
   Computation (IEEE-EC),* 18(2):226-243, April 2014.
   https://dx.doi.org/10.1109/TEVC.2013.2251885
4. Thomas Weise, Yan Chen, Xinlu Li, and Zhize Wu. Selecting a diverse set of
   benchmark instances from a tunable model problem for black-box discrete
   optimization algorithms. *Applied Soft Computing Journal (ASOC),*
   92:106269, June 2020. https://dx.doi.org/10.1016/j.asoc.2020.106269
5. Thomas Weise, Xinlu Li, Yan Chen, and Zhize Wu. Solving Job Shop Scheduling
   Problems Without Using a Bias for Good Solutions. In *Genetic and
   Evolutionary Computation Conference Companion (GECCO'21 Companion),*
   July 10-14, 2021, Lille, France. ACM, New York, NY, USA.
   ISBN 978-1-4503-8351-6. https://dx.doi.org/10.1145/3449726.3463124
6. Thomas Weise, Mingxu Wan, Ke Tang, and Xin Yao. Evolving Exact Integer
   Algorithms with Genetic Programming. In *Proceedings of the IEEE Congress
   on Evolutionary Computation (CEC'14), Proceedings of the 2014 World
   Congress on Computational Intelligence (WCCI'14),* pages 1816-1823,
   July 6-11, 2014, Beijing, China. Los Alamitos, CA, USA: IEEE Computer
   Society Press. ISBN: 978-1-4799-1488-3.
   https://dx.doi.org/10.1109/CEC.2014.6900292
"""


from collections import Counter
from typing import Final, cast

import numpy as np
from numpy.random import Generator

from moptipy.algorithms.so.fitness import Fitness, FRecord
from moptipy.api.objective import Objective, check_objective
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.nputils import DEFAULT_UNSIGNED_INT


class FFA(Fitness):
    """The frequency fitness assignment (FFA) process."""

    def __new__(cls, f: Objective) -> "FFA":
        """
        Create the frequency fitness assignment mapping.

        :param f: the objective function
        """
        check_objective(f)
        if f.is_always_integer():
            lb: Final[int | float] = f.lower_bound()
            ub: Final[int | float] = f.upper_bound()
            if isinstance(ub, int) and isinstance(lb, int) \
                    and ((ub - lb) <= 100_000_000):
                if 0 <= lb <= 10_000_000:
                    return _IntFFA1.__new__(_IntFFA1, cast(int, ub))
                return _IntFFA2.__new__(_IntFFA2, cast(int, lb),
                                        cast(int, ub))
        return _DictFFA.__new__(_DictFFA)

    def __str__(self):
        """
        Get the name (`"ffa"`) of this fitness assignment process.

        :return: the name of this process: `ffa`
        retval "ffa": always
        """
        return "ffa"


class _IntFFA1(FFA):
    """The internal FFA-based class."""

    #: the internal frequency table
    __h: np.ndarray

    def __new__(cls, ub: int) -> "_IntFFA1":
        """Initialize the pure integer FFA."""
        instance = object.__new__(_IntFFA1)
        instance.__h = np.zeros(ub + 1, dtype=DEFAULT_UNSIGNED_INT)
        return instance

    def assign_fitness(self, p: list[FRecord], random: Generator) -> None:
        """
        Assign the frequency fitness.

        :param p: the list of records
        :param random: ignored
        """
        h: Final[np.ndarray] = self.__h
        min_it: int = 9_223_372_036_854_775_808  # the minimum iteration index
        max_it: int = -1  # the maximum iteration index
        for r in p:
            it: int = r.it  # get iteration index from record
            if it < min_it:  # update minimum iteration index
                min_it = it
            if it > max_it:  # update maximum iteration index
                max_it = it
        if max_it <= 0:  # in the first iteration, all elements get the
            for r in p:  # same fitness: 0
                r.fitness = 0
            return
        for r in p:
            h[r.f] += 1  # type: ignore
        it_range: Final[int] = max_it - min_it + 1  # range of it index
        for r in p:
            r.fitness = ((int(h[r.f]) * it_range)  # type: ignore
                         + max_it - r.it)

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log all parameters of this component as key-value pairs.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        logger.key_value("ub", len(self.__h) - 1)

    def initialize(self) -> None:
        """Initialize the algorithm."""
        super().initialize()
        self.__h.fill(0)


class _IntFFA2(FFA):
    """The internal FFA-based class."""

    #: the internal frequency table
    __h: np.ndarray
    #: the lower bound
    __lb: int

    def __new__(cls, lb: int, ub: int) -> "_IntFFA2":
        """Initialize the pure integer FFA."""
        instance = object.__new__(_IntFFA2)
        instance.__h = np.zeros(ub - lb + 1, dtype=DEFAULT_UNSIGNED_INT)
        instance.__lb = lb
        return instance

    def assign_fitness(self, p: list[FRecord], random: Generator) -> None:
        """
        Assign the frequency fitness.

        :param p: the list of records
        :param random: ignored
        """
        h: Final[np.ndarray] = self.__h
        lb: Final[int] = self.__lb

        min_it: int = 9_223_372_036_854_775_808  # the minimum iteration index
        max_it: int = -1  # the maximum iteration index
        for r in p:
            it: int = r.it  # get iteration index from record
            if it < min_it:  # update minimum iteration index
                min_it = it
            if it > max_it:  # update maximum iteration index
                max_it = it
        if max_it <= 0:  # in the first iteration, all elements get the
            for r in p:  # same fitness: 0
                r.fitness = 0
            return
        for r in p:
            h[r.f - lb] += 1  # type: ignore
        it_range: Final[int] = max_it - min_it + 1  # range of it index
        for r in p:
            r.fitness = ((int(h[r.f - lb]) * it_range)  # type: ignore
                         + max_it - r.it)

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log all parameters of this component as key-value pairs.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        logger.key_value("lb", self.__lb)
        logger.key_value("ub", len(self.__h) + self.__lb - 1)

    def initialize(self) -> None:
        """Initialize the algorithm."""
        super().initialize()
        self.__h.fill(0)


class _DictFFA(FFA):
    """The internal FFA dict-based class."""

    #: the internal frequency table
    __h: Counter

    def __new__(cls) -> "_DictFFA":
        """Initialize the pure integer FFA."""
        instance = object.__new__(_DictFFA)
        instance.__h = Counter()
        return instance

    def assign_fitness(self, p: list[FRecord], random: Generator) -> None:
        """
        Assign the frequency fitness.

        :param p: the list of records
        :param random: ignored
        """
        h: Final[Counter[int | float]] = self.__h

        min_it: int = 9_223_372_036_854_775_808  # the minimum iteration index
        max_it: int = -1  # the maximum iteration index
        for r in p:
            it: int = r.it  # get iteration index from record
            if it < min_it:  # update minimum iteration index
                min_it = it
            if it > max_it:  # update maximum iteration index
                max_it = it
        if max_it <= 0:  # in the first iteration, all elements get the
            for r in p:  # same fitness: 0
                r.fitness = 0
            return
        for r in p:
            h[r.f] += 1
        it_range: Final[int] = max_it - min_it + 1  # range of it index
        for r in p:
            r.fitness = (h[r.f] * it_range) + max_it - r.it

    def initialize(self) -> None:
        """Initialize the algorithm."""
        super().initialize()
        self.__h.clear()
